#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import unittest
import warnings
import json
import io

import numpy as np
import numpy.testing as np_test
import networkx as nx

from pgmpy.readwrite import ProbModelXMLReader, ProbModelXMLWriter, get_probmodel_data
from pgmpy.models import BayesianModel
from pgmpy.factors.discrete import TabularCPD

try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        try:
            import xml.etree.ElementTree as etree

            print("running with ElementTree on Python 2.5+")
        except ImportError:
            warnings.warn("Failed to import ElementTree from any known place")


class TestProbModelXMLReaderString(unittest.TestCase):
    def setUp(self):
        string = """<ProbModelXML formatVersion="1.0">
      <ProbNet type="BayesianNetwork">
        <AdditionalConstraints>
            <Constraint name="MaxNumParents">
                <Argument name="numParents" value="5" />
            </Constraint>
        </AdditionalConstraints>
        <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
        <Language>English</Language>
        <AdditionalProperties>
            <Property name="elvira.title" value="X ray result"/>
        </AdditionalProperties>
        <DecisionCriteria>
            <Criterion name="cost">
                <AdditionalProperties/>
            </Criterion>
            <Criterion name="effectiveness">
                <AdditionalProperties/>
            </Criterion>
        </DecisionCriteria>
        <Variables>
            <Variable name="intelligence" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="smart"><AdditionalProperties /></State>
                    <State name="dumb"><AdditionalProperties /></State>
                </States>
            </Variable>
            <Variable name="difficulty" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="difficult"><AdditionalProperties /></State>
                    <State name="easy"><AdditionalProperties /></State>
                </States>
            </Variable>
        </Variables>
        <Links>
            <Link directed="1">
                <Variable name="difficulty" />
                <Variable name="grade" />
                <Comment>Directed Edge from difficulty to grade</Comment>
                <Label>diff_to_grad</Label>
                <AdditionalProperties />
            </Link>
            <Link directed="1">
                <Variable name="intelligence" />
                <Variable name="grade" />
                <Comment>Directed Edge from intelligence to grade</Comment>
                <Label>intel_to_grad</Label>
                <AdditionalProperties />
            </Link>
            <Link directed="1">
                <Variable name="intelligence" />
                <Variable name="SAT" />
                <Comment>Directed Edge from intelligence to SAT</Comment>
                <Label>intel_to_sat</Label>
                <AdditionalProperties />
            </Link>
            <Link directed="1">
                <Variable name="grade" />
                <Variable name="recommendation_letter" />
                <Comment>Directed Edge from grade to recommendation_letter</Comment>
                <Label>grad_to_reco</Label>
                <AdditionalProperties />
            </Link>
        </Links>
        <Potentials>
          <Potential type="Tree/ADD" role="Utility">
                <UtilityVariable name="U1" />
                <Variables>
                    <Variable name="D0"/>
                    <Variable name="D1"/>
                    <Variable name="C0"/>
                    <Variable name="C1"/>
                 </Variables>
                <TopVariable name="D0"/>
                <Branches>
                   <Branch>
                      <States><State name="no"/></States>
                      <Potential type="Tree/ADD">
                         <TopVariable name="C1"/>
                         <Branches>
                              <Branch>
                                  <Thresholds>
                                       <Threshold value="–Infinity"/>
                                       <Threshold value="0" belongsTo="Left"/>
                                  </Thresholds>
                                  <Potential type="MixtureOfExponentials">
                                      <Variables>
                                         <Variable name="C0"/>
                                         <Variable name="C1"/>
                                      </Variables>
                                      <Subpotentials>
                                         <Potential type="Exponential">
                                             <Potential type="Table">
                                                  <Values>3</Values>
                                              </Potential>
                                          </Potential>
                                          <Potential type="Exponential">
                                              <Potential type="Table">
                                                  <Values>–1</Values>
                                              </Potential>
                                              <NumericVariables>
                                                  <Variable name="C0"/>
                                                  <Variable name="C1"/>
                                              </NumericVariables>
                                              <Coefficients>4 –1</Coefficients>
                                           </Potential>
                                      </Subpotentials>
                                 </Potential>
                              </Branch>
                              <Branch>
                                    <Thresholds>
                                        <Threshold value="0" belongsTo="Left" />
                                        <Threshold value="+Infinity" />
                                     </Thresholds>
                                     <Potential type="MixtureOfExponentials">
                                        <Variables>
                                              <Variable name="C1"/>
                                              <Variable name="D1"/>
                                        </Variables>
                                        <Subpotentials>
                                              <Potential type="Exponential">
                                                 <Potential type="Table">
                                                      <Variables>
                                                          <Variable name="D1"/>
                                                      </Variables>
                                                      <Values>10  5</Values>
                                                 </Potential>
                                                 <NumericVariables>
                                                    <Variable name="C1"/>
                                                 </NumericVariables>
                                                 <Coefficients>0.25</Coefficients>
                                              </Potential>
                                         </Subpotentials>
                                     </Potential>
                              </Branch>
                         </Branches>
                      </Potential>
                   </Branch>
                  <Branch>
                        <States>
                            <State name="yes"/>
                         </States>
                         <Potential type="MixtureOfExponentials">
                                <Variables>
                                    <Variable name="C0"/>
                                </Variables>
                                <Subpotentials>
                                    <Potential type="Exponential">
                                        <Potential type="Table">
                                             <Values>0.3</Values>
                                        </Potential>
                                        <NumericVariables>
                                             <Variable name="C0"/>
                                        </NumericVariables>
                                        <Coefficients>1</Coefficients>
                                    </Potential>
                                   <Potential type="Exponential">
                                        <Potential type="Table">
                                             <Values>0.7</Values>
                                        </Potential>
                                   </Potential>
                                </Subpotentials>
                         </Potential>
                  </Branch>
                </Branches>
          </Potential>
      </Potentials>
    </ProbNet>
    <Policies />
    <InferenceOptions />
</ProbModelXML>
"""
        self.maxDiff = None
        self.reader_string = ProbModelXMLReader(string=string)
        self.reader_file = ProbModelXMLReader(path=io.StringIO(string))

    def test_comment(self):
        comment_expected = (
            "Student example model from Probabilistic Graphical Models: "
            "Principles and Techniques by Daphne Koller"
        )
        self.assertEqual(self.reader_string.probnet["Comment"], comment_expected)
        self.assertEqual(self.reader_file.probnet["Comment"], comment_expected)

    def test_variables(self):
        variables = {
            "difficulty": {
                "Comment": None,
                "Coordinates": {},
                "role": "Chance",
                "type": "FiniteState",
                "States": {"difficult": {}, "easy": {}},
            },
            "intelligence": {
                "Comment": None,
                "Coordinates": {},
                "role": "Chance",
                "type": "FiniteState",
                "States": {"smart": {}, "dumb": {}},
            },
        }
        self.assertDictEqual(self.reader_string.probnet["Variables"], variables)
        self.assertDictEqual(self.reader_file.probnet["Variables"], variables)

    def test_edges(self):
        edge = {
            ("grade", "recommendation_letter"): {
                "directed": "1",
                "Comment": "Directed Edge from grade to recommendation_letter",
                "Label": "grad_to_reco",
            },
            ("intelligence", "grade"): {
                "directed": "1",
                "Comment": "Directed Edge from intelligence to grade",
                "Label": "intel_to_grad",
            },
            ("difficulty", "grade"): {
                "directed": "1",
                "Comment": "Directed Edge from difficulty to grade",
                "Label": "diff_to_grad",
            },
            ("intelligence", "SAT"): {
                "directed": "1",
                "Comment": "Directed Edge from intelligence to SAT",
                "Label": "intel_to_sat",
            },
        }
        self.assertDictEqual(self.reader_string.probnet["edges"], edge)
        self.assertDictEqual(self.reader_file.probnet["edges"], edge)

    def test_additionalconstraints(self):
        additionalconstraints_expected = {"MaxNumParents": {"numParents": "5"}}
        self.assertDictEqual(
            self.reader_string.probnet["AdditionalConstraints"],
            additionalconstraints_expected,
        )
        self.assertDictEqual(
            self.reader_file.probnet["AdditionalConstraints"],
            additionalconstraints_expected,
        )

    def test_additionalproperties(self):
        additionalproperties_expected = {"elvira.title": "X ray result"}
        self.assertDictEqual(
            self.reader_string.probnet["AdditionalProperties"],
            additionalproperties_expected,
        )
        self.assertDictEqual(
            self.reader_file.probnet["AdditionalProperties"],
            additionalproperties_expected,
        )

    def test_decisioncriteria(self):
        decisioncriteria_expected = {"effectiveness": {}, "cost": {}}
        self.assertDictEqual(
            self.reader_string.probnet["DecisionCriteria"], decisioncriteria_expected
        )
        self.assertDictEqual(
            self.reader_file.probnet["DecisionCriteria"], decisioncriteria_expected
        )

    def test_potential(self):
        potential_expected = [
            {
                "role": "Utility",
                "Variables": {"D0": ["D1", "C0", "C1"]},
                "type": "Tree/ADD",
                "UtilityVaribale": "U1",
                "Branches": [
                    {
                        "Potential": {
                            "type": "Tree/ADD",
                            "Branches": [
                                {
                                    "Thresholds": [
                                        {"value": u"–Infinity"},
                                        {"value": "0", "belongsTo": "Left"},
                                    ],
                                    "Potential": {
                                        "Subpotentials": [
                                            {
                                                "Potential": {
                                                    "type": "Table",
                                                    "Values": "3",
                                                },
                                                "type": "Exponential",
                                            },
                                            {
                                                "NumericVariables": ["C0", "C1"],
                                                "Potential": {
                                                    "type": "Table",
                                                    "Values": u"–1",
                                                },
                                                "Coefficients": u"4 –1",
                                                "type": "Exponential",
                                            },
                                        ],
                                        "Variables": {"C0": ["C1"]},
                                        "type": "MixtureOfExponentials",
                                    },
                                },
                                {
                                    "Thresholds": [
                                        {"value": "0", "belongsTo": "Left"},
                                        {"value": "+Infinity"},
                                    ],
                                    "Potential": {
                                        "Subpotentials": [
                                            {
                                                "NumericVariables": ["C1"],
                                                "Potential": {
                                                    "Variables": {"D1": []},
                                                    "type": "Table",
                                                    "Values": "10  5",
                                                },
                                                "Coefficients": "0.25",
                                                "type": "Exponential",
                                            }
                                        ],
                                        "Variables": {"C1": ["D1"]},
                                        "type": "MixtureOfExponentials",
                                    },
                                },
                            ],
                            "TopVariable": "C1",
                        },
                        "States": [{"name": "no"}],
                    },
                    {
                        "Potential": {
                            "Subpotentials": [
                                {
                                    "NumericVariables": ["C0"],
                                    "Potential": {"type": "Table", "Values": "0.3"},
                                    "Coefficients": "1",
                                    "type": "Exponential",
                                },
                                {
                                    "Potential": {"type": "Table", "Values": "0.7"},
                                    "type": "Exponential",
                                },
                            ],
                            "Variables": {"C0": []},
                            "type": "MixtureOfExponentials",
                        },
                        "States": [{"name": "yes"}],
                    },
                ],
                "TopVariable": "D0",
            }
        ]

        self.assertListEqual(
            self.reader_string.probnet["Potentials"], potential_expected
        )
        self.assertListEqual(self.reader_file.probnet["Potentials"], potential_expected)

    def test_get_model(self):
        string = """<ProbModelXML formatVersion="0.2.0">
  <ProbNet type="BayesianNetwork">
    <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
    <Variables>
      <Variable name="X-ray" type="finiteStates" role="chance" isInput="false">
        <Comment><![CDATA[Indica si el test de rayos X ha sido positivo]]></Comment>
        <Coordinates x="252" y="322" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="X" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Bronchitis" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="698" y="181" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="B" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Dyspnea" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="533" y="321" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="D" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="VisitToAsia" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="290" y="58" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="A" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Smoker" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="568" y="52" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="S" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="LungCancer" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="421" y="152" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="L" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Tuberculosis" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="201" y="150" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="T" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="TuberculosisOrCancer" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="336" y="238" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="E" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="true">
        <Variable name="Bronchitis" />
        <Variable name="Dyspnea" />
      </Link>
      <Link directed="true">
        <Variable name="VisitToAsia" />
        <Variable name="Tuberculosis" />
      </Link>
      <Link directed="true">
        <Variable name="Smoker" />
        <Variable name="Bronchitis" />
      </Link>
      <Link directed="true">
        <Variable name="Smoker" />
        <Variable name="LungCancer" />
      </Link>
      <Link directed="true">
        <Variable name="LungCancer" />
        <Variable name="TuberculosisOrCancer" />
      </Link>
      <Link directed="true">
        <Variable name="Tuberculosis" />
        <Variable name="TuberculosisOrCancer" />
      </Link>
      <Link directed="true">
        <Variable name="TuberculosisOrCancer" />
        <Variable name="Dyspnea" />
      </Link>
      <Link directed="true">
        <Variable name="TuberculosisOrCancer" />
        <Variable name="X-ray" />
      </Link>
    </Links>
    <Potentials>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="X-ray" />
          <Variable name="TuberculosisOrCancer" />
        </Variables>
        <Values>0.95 0.05 0.02 0.98</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Bronchitis" />
          <Variable name="Smoker" />
        </Variables>
        <Values>0.7 0.3 0.4 0.6</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Dyspnea" />
          <Variable name="TuberculosisOrCancer" />
          <Variable name="Bronchitis" />
        </Variables>
        <Values>0.9 0.1 0.3 0.7 0.2 0.8 0.1 0.9</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="VisitToAsia" />
        </Variables>
        <Values>0.99 0.01</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Smoker" />
        </Variables>
        <Values>0.5 0.5</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="LungCancer" />
          <Variable name="Smoker" />
        </Variables>
        <Values>0.99 0.01 0.9 0.1</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Tuberculosis" />
          <Variable name="VisitToAsia" />
        </Variables>
        <Values>0.99 0.01 0.95 0.05</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="TuberculosisOrCancer" />
          <Variable name="LungCancer" />
          <Variable name="Tuberculosis" />
        </Variables>
        <Values>1.0 0.0 0.0 1.0 0.0 1.0 0.0 1.0</Values>
      </Potential>
    </Potentials>
    <AdditionalProperties>
      <Property name="VisualPrecision" value="0.0" />
      <Property name="KindOfGraph" value="directed" />
      <Property name="WhenChanged" value="19/08/99" />
      <Property name="Version" value="1.0" />
      <Property name="WhoChanged" value="Jose A. Gamez" />
    </AdditionalProperties>
  </ProbNet>
</ProbModelXML>"""
        self.maxDiff = None
        self.reader = ProbModelXMLReader(string=string)
        model = self.reader.get_model()
        edges_expected = [
            ("VisitToAsia", "Tuberculosis"),
            ("LungCancer", "TuberculosisOrCancer"),
            ("Smoker", "LungCancer"),
            ("Smoker", "Bronchitis"),
            ("Tuberculosis", "TuberculosisOrCancer"),
            ("Bronchitis", "Dyspnea"),
            ("TuberculosisOrCancer", "Dyspnea"),
            ("TuberculosisOrCancer", "X-ray"),
        ]
        node_expected = {
            "Smoker": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "52", "x": "568"},
                "AdditionalProperties": {"Title": "S", "Relevance": "7.0"},
                "weight": None,
            },
            "Bronchitis": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "181", "x": "698"},
                "AdditionalProperties": {"Title": "B", "Relevance": "7.0"},
                "weight": None,
            },
            "VisitToAsia": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "58", "x": "290"},
                "AdditionalProperties": {"Title": "A", "Relevance": "7.0"},
                "weight": None,
            },
            "Tuberculosis": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "150", "x": "201"},
                "AdditionalProperties": {"Title": "T", "Relevance": "7.0"},
                "weight": None,
            },
            "X-ray": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "AdditionalProperties": {"Title": "X", "Relevance": "7.0"},
                "Coordinates": {"y": "322", "x": "252"},
                "Comment": "Indica si el test de rayos X ha sido positivo",
                "type": "finiteStates",
                "weight": None,
            },
            "Dyspnea": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "321", "x": "533"},
                "AdditionalProperties": {"Title": "D", "Relevance": "7.0"},
                "weight": None,
            },
            "TuberculosisOrCancer": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "238", "x": "336"},
                "AdditionalProperties": {"Title": "E", "Relevance": "7.0"},
                "weight": None,
            },
            "LungCancer": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "152", "x": "421"},
                "AdditionalProperties": {"Title": "L", "Relevance": "7.0"},
                "weight": None,
            },
        }
        edge_expected = {
            "LungCancer": {
                "TuberculosisOrCancer": {"weight": None, "directed": "true"}
            },
            "Smoker": {
                "LungCancer": {"weight": None, "directed": "true"},
                "Bronchitis": {"weight": None, "directed": "true"},
            },
            "Dyspnea": {},
            "X-ray": {},
            "VisitToAsia": {"Tuberculosis": {"weight": None, "directed": "true"}},
            "TuberculosisOrCancer": {
                "X-ray": {"weight": None, "directed": "true"},
                "Dyspnea": {"weight": None, "directed": "true"},
            },
            "Bronchitis": {"Dyspnea": {"weight": None, "directed": "true"}},
            "Tuberculosis": {
                "TuberculosisOrCancer": {"weight": None, "directed": "true"}
            },
        }

        cpds_expected = [
            np.array([[0.95, 0.05], [0.02, 0.98]]),
            np.array([[0.7, 0.3], [0.4, 0.6]]),
            np.array([[0.9, 0.1, 0.3, 0.7], [0.2, 0.8, 0.1, 0.9]]),
            np.array([[0.99], [0.01]]),
            np.array([[0.5], [0.5]]),
            np.array([[0.99, 0.01], [0.9, 0.1]]),
            np.array([[0.99, 0.01], [0.95, 0.05]]),
            np.array([[1, 0, 0, 1], [0, 1, 0, 1]]),
        ]
        for cpd_index in range(0, len(cpds_expected)):
            np_test.assert_array_equal(
                model.get_cpds()[cpd_index].get_values(), cpds_expected[cpd_index]
            )
        self.assertDictEqual(dict(model.nodes), node_expected)
        if nx.__version__.startswith("1"):
            self.assertDictEqual(model.edge, edge_expected)
        else:
            self.assertDictEqual(dict(model.adj), edge_expected)
        self.assertListEqual(sorted(model.edges()), sorted(edges_expected))


class TestProbModelXMLWriter(unittest.TestCase):
    def setUp(self):
        self.model_data = {
            "probnet": {
                "type": "BayesianNetwork",
                "Language": "English",
                "AdditionalConstraints": {"MaxNumParents": {"numParents": "5"}},
                "AdditionalProperties": {"elvira.title": "X ray result"},
                "DecisionCriteria": {"effectiveness": {}, "cost": {}},
                "Variables": {
                    "difficulty": {
                        "type": "FiniteState",
                        "role": "Chance",
                        "States": {"difficult": {}, "easy": {}},
                        "Comment": None,
                        "Coordinates": {},
                    },
                    "intelligence": {
                        "type": "FiniteState",
                        "role": "Chance",
                        "States": {"smart": {}, "dumb": {}},
                        "Comment": None,
                        "Coordinates": {},
                    },
                },
                "Comment": "Student example model from Probabilistic Graphical Models: "
                "Principles and Techniques by Daphne Koller",
                "edges": {
                    ("difficulty", "grade"): {
                        "directed": "1",
                        "Label": "diff_to_grad",
                        "Comment": "Directed Edge from difficulty to grade",
                    },
                    ("intelligence", "grade"): {
                        "directed": "1",
                        "Label": "intel_to_grad",
                        "Comment": "Directed Edge from intelligence to grade",
                    },
                    ("intelligence", "SAT"): {
                        "directed": "1",
                        "Label": "intel_to_sat",
                        "Comment": "Directed Edge from intelligence to SAT",
                    },
                    ("grade", "recommendation_letter"): {
                        "directed": "1",
                        "Label": "grad_to_reco",
                        "Comment": "Directed Edge from grade to recommendation_letter",
                    },
                },
                "Potentials": [
                    {
                        "role": "Utility",
                        "Variables": {"D0": ["D1", "C0", "C1"]},
                        "type": "Tree/ADD",
                        "UtilityVaribale": "U1",
                        "Branches": [
                            {
                                "Potential": {
                                    "type": "Tree/ADD",
                                    "Branches": [
                                        {
                                            "Thresholds": [
                                                {"value": "-Infinity"},
                                                {"value": "0", "belongsTo": "Left"},
                                            ],
                                            "Potential": {
                                                "Subpotentials": [
                                                    {
                                                        "Potential": {
                                                            "type": "Table",
                                                            "Values": "3",
                                                        },
                                                        "type": "Exponential",
                                                    },
                                                    {
                                                        "NumericVariables": [
                                                            "C0",
                                                            "C1",
                                                        ],
                                                        "Potential": {
                                                            "type": "Table",
                                                            "Values": "-1",
                                                        },
                                                        "Coefficients": "4 -1",
                                                        "type": "Exponential",
                                                    },
                                                ],
                                                "Variables": {"C0": ["C1"]},
                                                "type": "MixtureOfExponentials",
                                            },
                                        },
                                        {
                                            "Thresholds": [
                                                {"value": "0", "belongsTo": "Left"},
                                                {"value": "+Infinity"},
                                            ],
                                            "Potential": {
                                                "Subpotentials": [
                                                    {
                                                        "NumericVariables": ["C1"],
                                                        "Potential": {
                                                            "Variables": {"D1": []},
                                                            "type": "Table",
                                                            "Values": "10  5",
                                                        },
                                                        "Coefficients": "0.25",
                                                        "type": "Exponential",
                                                    }
                                                ],
                                                "Variables": {"C1": ["D1"]},
                                                "type": "MixtureOfExponentials",
                                            },
                                        },
                                    ],
                                    "TopVariable": "C1",
                                },
                                "States": [{"name": "no"}],
                            },
                            {
                                "Potential": {
                                    "Subpotentials": [
                                        {
                                            "NumericVariables": ["C0"],
                                            "Potential": {
                                                "type": "Table",
                                                "Values": "0.3",
                                            },
                                            "Coefficients": "1",
                                            "type": "Exponential",
                                        },
                                        {
                                            "Potential": {
                                                "type": "Table",
                                                "Values": "0.7",
                                            },
                                            "type": "Exponential",
                                        },
                                    ],
                                    "Variables": {"C0": []},
                                    "type": "MixtureOfExponentials",
                                },
                                "States": [{"name": "yes"}],
                            },
                        ],
                        "TopVariable": "D0",
                    }
                ],
            }
        }

        self.maxDiff = None
        self.writer = ProbModelXMLWriter(model_data=self.model_data)

    def test_file(self):
        self.expected_xml = etree.XML(
            """<ProbModelXML formatVersion="1.0">
  <ProbNet type="BayesianNetwork">
    <Variables>
      <Variable name="difficulty" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="difficult">
            <AdditionalProperties/>
          </State>
          <State name="easy">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="intelligence" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="dumb">
            <AdditionalProperties/>
          </State>
          <State name="smart">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="1" var1="difficulty" var2="grade">
        <Comment>Directed Edge from difficulty to grade</Comment>
        <Label>diff_to_grad</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="grade" var2="recommendation_letter">
        <Comment>Directed Edge from grade to recommendation_letter</Comment>
        <Label>grad_to_reco</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="SAT">
        <Comment>Directed Edge from intelligence to SAT</Comment>
        <Label>intel_to_sat</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="grade">
        <Comment>Directed Edge from intelligence to grade</Comment>
        <Label>intel_to_grad</Label>
        <AdditionalProperties/>
      </Link>
    </Links>
    <Potentials>
      <Potential role="Utility" type="Tree/ADD">
        <Variables>
          <Variable name="D0"/>
          <Variable name="C0"/>
          <Variable name="C1"/>
          <Variable name="D1"/>
        </Variables>
        <TopVariable name="D0"/>
        <Branches>
          <Branch>
            <States>
              <State name="no"/>
            </States>
            <Potential type="Tree/ADD">
              <TopVariable name="C1"/>
              <Branches>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C0"/>
                      <Variable name="C1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Potential type="Table">
                          <Values>3</Values>
                        </Potential>
                      </Potential>
                      <Potential type="Exponential">
                        <Coefficients>4 -1</Coefficients>
                        <Potential type="Table">
                          <Values>-1</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C0"/>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold value="-Infinity"/>
                    <Threshold belongsTo="Left" value="0"/>
                  </Thresholds>
                </Branch>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C1"/>
                      <Variable name="D1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Coefficients>0.25</Coefficients>
                        <Potential type="Table">
                          <Variables>
                            <Variable name="D1"/>
                          </Variables>
                          <Values>10  5</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold belongsTo="Left" value="0"/>
                    <Threshold value="+Infinity"/>
                  </Thresholds>
                </Branch>
              </Branches>
            </Potential>
          </Branch>
          <Branch>
            <States>
              <State name="yes"/>
            </States>
            <Potential type="MixtureOfExponentials">
              <Variables>
                <Variable name="C0"/>
              </Variables>
              <Subpotentials>
                <Potential type="Exponential">
                  <Coefficients>1</Coefficients>
                  <Potential type="Table">
                    <Values>0.3</Values>
                  </Potential>
                  <NumericVariables>
                    <Variable name="C0"/>
                  </NumericVariables>
                </Potential>
                <Potential type="Exponential">
                  <Potential type="Table">
                    <Values>0.7</Values>
                  </Potential>
                </Potential>
              </Subpotentials>
            </Potential>
          </Branch>
        </Branches>
      </Potential>
    </Potentials>
    <AdditionalConstraints>
      <Constraint name="MaxNumParents">
        <Argument name="numParents" value="5"/>
      </Constraint>
    </AdditionalConstraints>
    <Language>English</Language>
    <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
  </ProbNet>
  <AdditionalProperties>
    <Property name="elvira.title" value="X ray result"/>
  </AdditionalProperties>
  <DecisionCriteria>
    <Criterion name="cost">
      <AdditionalProperties/>
    </Criterion>
    <Criterion name="effectiveness">
      <AdditionalProperties/>
    </Criterion>
  </DecisionCriteria>
</ProbModelXML>"""
        )
        self.assertEqual(
            str(self.writer.__str__()[:-1]), str(etree.tostring(self.expected_xml))
        )

    def test_write_file(self):
        self.expected_xml = etree.XML(
            """<ProbModelXML formatVersion="1.0">
  <ProbNet type="BayesianNetwork">
    <Variables>
      <Variable name="difficulty" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="difficult">
            <AdditionalProperties/>
          </State>
          <State name="easy">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="intelligence" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="dumb">
            <AdditionalProperties/>
          </State>
          <State name="smart">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="1" var1="difficulty" var2="grade">
        <Comment>Directed Edge from difficulty to grade</Comment>
        <Label>diff_to_grad</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="grade" var2="recommendation_letter">
        <Comment>Directed Edge from grade to recommendation_letter</Comment>
        <Label>grad_to_reco</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="SAT">
        <Comment>Directed Edge from intelligence to SAT</Comment>
        <Label>intel_to_sat</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="grade">
        <Comment>Directed Edge from intelligence to grade</Comment>
        <Label>intel_to_grad</Label>
        <AdditionalProperties/>
      </Link>
    </Links>
    <Potentials>
      <Potential role="Utility" type="Tree/ADD">
        <Variables>
          <Variable name="D0"/>
          <Variable name="C0"/>
          <Variable name="C1"/>
          <Variable name="D1"/>
        </Variables>
        <TopVariable name="D0"/>
        <Branches>
          <Branch>
            <States>
              <State name="no"/>
            </States>
            <Potential type="Tree/ADD">
              <TopVariable name="C1"/>
              <Branches>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C0"/>
                      <Variable name="C1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Potential type="Table">
                          <Values>3</Values>
                        </Potential>
                      </Potential>
                      <Potential type="Exponential">
                        <Coefficients>4 -1</Coefficients>
                        <Potential type="Table">
                          <Values>-1</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C0"/>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold value="-Infinity"/>
                    <Threshold belongsTo="Left" value="0"/>
                  </Thresholds>
                </Branch>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C1"/>
                      <Variable name="D1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Coefficients>0.25</Coefficients>
                        <Potential type="Table">
                          <Variables>
                            <Variable name="D1"/>
                          </Variables>
                          <Values>10  5</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold belongsTo="Left" value="0"/>
                    <Threshold value="+Infinity"/>
                  </Thresholds>
                </Branch>
              </Branches>
            </Potential>
          </Branch>
          <Branch>
            <States>
              <State name="yes"/>
            </States>
            <Potential type="MixtureOfExponentials">
              <Variables>
                <Variable name="C0"/>
              </Variables>
              <Subpotentials>
                <Potential type="Exponential">
                  <Coefficients>1</Coefficients>
                  <Potential type="Table">
                    <Values>0.3</Values>
                  </Potential>
                  <NumericVariables>
                    <Variable name="C0"/>
                  </NumericVariables>
                </Potential>
                <Potential type="Exponential">
                  <Potential type="Table">
                    <Values>0.7</Values>
                  </Potential>
                </Potential>
              </Subpotentials>
            </Potential>
          </Branch>
        </Branches>
      </Potential>
    </Potentials>
    <AdditionalConstraints>
      <Constraint name="MaxNumParents">
        <Argument name="numParents" value="5"/>
      </Constraint>
    </AdditionalConstraints>
    <Language>English</Language>
    <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
  </ProbNet>
  <AdditionalProperties>
    <Property name="elvira.title" value="X ray result"/>
  </AdditionalProperties>
  <DecisionCriteria>
    <Criterion name="cost">
      <AdditionalProperties/>
    </Criterion>
    <Criterion name="effectiveness">
      <AdditionalProperties/>
    </Criterion>
  </DecisionCriteria>
</ProbModelXML>"""
        )
        self.writer.write_file("test_xml.pgmx")
        with open("test_xml.pgmx", "r") as myfile:
            data = myfile.read()
        self.assertEqual(
            str(self.writer.__str__()[:-1]), str(etree.tostring(self.expected_xml))
        )
        self.assertEqual(
            str(data), str(etree.tostring(self.expected_xml).decode("utf-8"))
        )

    def tearDown(self):
        try:
            os.remove("test_xml.pgmx")
        except OSError:
            pass


class TestProbModelXMLmethods(unittest.TestCase):
    def setUp(self):
        variables = [
            "VisitToAsia",
            "Tuberculosis",
            "Smoker",
            "LungCancer",
            "Bronchitis",
            "Dyspnea",
            "TuberculosisOrCancer",
            "X-ray",
            "Sinus",
        ]

        edges_list = [
            ("VisitToAsia", "Tuberculosis"),
            ("LungCancer", "TuberculosisOrCancer"),
            ("Smoker", "LungCancer"),
            ("Smoker", "Bronchitis"),
            ("Tuberculosis", "TuberculosisOrCancer"),
            ("Bronchitis", "Dyspnea"),
            ("TuberculosisOrCancer", "Dyspnea"),
            ("TuberculosisOrCancer", "X-ray"),
        ]
        nodes = {
            "Smoker": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "52", "x": "568"},
                "AdditionalProperties": {"Title": "S", "Relevance": "7.0"},
            },
            "Bronchitis": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "181", "x": "698"},
                "AdditionalProperties": {"Title": "B", "Relevance": "7.0"},
            },
            "VisitToAsia": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "58", "x": "290"},
                "AdditionalProperties": {"Title": "A", "Relevance": "7.0"},
            },
            "Tuberculosis": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "150", "x": "201"},
                "AdditionalProperties": {"Title": "T", "Relevance": "7.0"},
            },
            "X-ray": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "AdditionalProperties": {"Title": "X", "Relevance": "7.0"},
                "Coordinates": {"y": "322", "x": "252"},
                "Comment": "Indica si el test de rayos X ha sido positivo",
                "type": "finiteStates",
            },
            "Dyspnea": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "321", "x": "533"},
                "AdditionalProperties": {"Title": "D", "Relevance": "7.0"},
            },
            "TuberculosisOrCancer": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "238", "x": "336"},
                "AdditionalProperties": {"Title": "E", "Relevance": "7.0"},
            },
            "LungCancer": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "152", "x": "421"},
                "AdditionalProperties": {"Title": "L", "Relevance": "7.0"},
            },
            "Sinus": {
                "States": {"no": {}, "yes": {}},
                "role": "chance",
                "type": "finiteStates",
                "Coordinates": {"y": "200", "x": "100"},
                "AdditionalProperties": {"Title": "S", "Relevance": "7.0"},
            },
        }
        edges = {
            "LungCancer": {"TuberculosisOrCancer": {"directed": "true"}},
            "Smoker": {
                "LungCancer": {"directed": "true"},
                "Bronchitis": {"directed": "true"},
            },
            "Dyspnea": {},
            "X-ray": {},
            "VisitToAsia": {"Tuberculosis": {"directed": "true"}},
            "TuberculosisOrCancer": {
                "X-ray": {"directed": "true"},
                "Dyspnea": {"directed": "true"},
            },
            "Bronchitis": {"Dyspnea": {"directed": "true"}},
            "Tuberculosis": {"TuberculosisOrCancer": {"directed": "true"}},
        }

        cpds = [
            {
                "Values": np.array([[0.95, 0.05], [0.02, 0.98]]),
                "Variables": {"X-ray": ["TuberculosisOrCancer"]},
            },
            {
                "Values": np.array([[0.7, 0.3], [0.4, 0.6]]),
                "Variables": {"Bronchitis": ["Smoker"]},
            },
            {
                "Values": np.array([[0.9, 0.1, 0.3, 0.7], [0.2, 0.8, 0.1, 0.9]]),
                "Variables": {"Dyspnea": ["TuberculosisOrCancer", "Bronchitis"]},
            },
            {"Values": np.array([[0.99], [0.01]]), "Variables": {"VisitToAsia": []}},
            {"Values": np.array([[0.5], [0.5]]), "Variables": {"Smoker": []}},
            {
                "Values": np.array([[0.99, 0.01], [0.9, 0.1]]),
                "Variables": {"LungCancer": ["Smoker"]},
            },
            {
                "Values": np.array([[0.99, 0.01], [0.95, 0.05]]),
                "Variables": {"Tuberculosis": ["VisitToAsia"]},
            },
            {
                "Values": np.array([[1, 0, 0, 1], [0, 1, 0, 1]]),
                "Variables": {"TuberculosisOrCancer": ["LungCancer", "Tuberculosis"]},
            },
            {"Values": np.array([[0.3], [0.7]]), "Variables": {"Sinus": []}},
        ]
        self.model = BayesianModel()
        self.model.add_nodes_from(variables)
        self.model.add_edges_from(edges_list)

        if nx.__version__.startswith("1"):
            for node in nodes:
                self.model.nodes[node] = nodes[node]
            for edge in edges:
                self.model.edge[edge] = edges[edge]
        else:
            for node in nodes:
                self.model._node[node] = nodes[node]
            for edge in edges:
                self.model._adj[edge] = edges[edge]

        tabular_cpds = []
        for cpd in cpds:
            var = list(cpd["Variables"].keys())[0]
            evidence = cpd["Variables"][var]
            values = cpd["Values"]
            states = len(nodes[var]["States"])
            evidence_card = [
                len(nodes[evidence_var]["States"]) for evidence_var in evidence
            ]
            tabular_cpds.append(
                TabularCPD(var, states, values, evidence, evidence_card)
            )
        self.maxDiff = None
        self.model.add_cpds(*tabular_cpds)

    def test_get_probmodel_data(self):
        model_data = get_probmodel_data(self.model)
        xmlfile = ProbModelXMLWriter(model_data)
        with open(
            "pgmpy/tests/test_readwrite/testdata/test_probmodelxml_data.json"
        ) as data_file:
            model_data_expected = json.load(data_file)
        xmlfile_expected = etree.XML(
            """<ProbModelXML formatVersion="1.0">
  <ProbNet type="BayesianNetwork">
    <Variables>
      <Variable name="Bronchitis" role="chance" type="finiteStates">
        <Coordinates x="698" y="181"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="B"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="Dyspnea" role="chance" type="finiteStates">
        <Coordinates x="533" y="321"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="D"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="LungCancer" role="chance" type="finiteStates">
        <Coordinates x="421" y="152"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="L"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="Sinus" role="chance" type="finiteStates">
        <Coordinates x="100" y="200"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="S"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="Smoker" role="chance" type="finiteStates">
        <Coordinates x="568" y="52"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="S"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="Tuberculosis" role="chance" type="finiteStates">
        <Coordinates x="201" y="150"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="T"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="TuberculosisOrCancer" role="chance" type="finiteStates">
        <Coordinates x="336" y="238"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="E"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="VisitToAsia" role="chance" type="finiteStates">
        <Coordinates x="290" y="58"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="A"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="X-ray" role="chance" type="finiteStates">
        <Comment>Indica si el test de rayos X ha sido positivo</Comment>
        <Coordinates x="252" y="322"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="X"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="true" var1="Bronchitis" var2="Dyspnea">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="LungCancer" var2="TuberculosisOrCancer">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="Smoker" var2="Bronchitis">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="Smoker" var2="LungCancer">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="Tuberculosis" var2="TuberculosisOrCancer">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="TuberculosisOrCancer" var2="Dyspnea">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="TuberculosisOrCancer" var2="X-ray">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="VisitToAsia" var2="Tuberculosis">
        <AdditionalProperties/>
      </Link>
    </Links>
    <Potentials>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="X-ray"/>
          <Variable name="TuberculosisOrCancer"/>
        </Variables>
        <Values>0.95 0.05 0.02 0.98 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Bronchitis"/>
          <Variable name="Smoker"/>
        </Variables>
        <Values>0.7 0.3 0.4 0.6 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Dyspnea"/>
          <Variable name="Bronchitis"/>
          <Variable name="TuberculosisOrCancer"/>
        </Variables>
        <Values>0.9 0.1 0.3 0.7 0.2 0.8 0.1 0.9 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="VisitToAsia"/>
        </Variables>
        <Values>0.99 0.01 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Smoker"/>
        </Variables>
        <Values>0.5 0.5 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="LungCancer"/>
          <Variable name="Smoker"/>
        </Variables>
        <Values>0.99 0.01 0.9 0.1 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Tuberculosis"/>
          <Variable name="VisitToAsia"/>
        </Variables>
        <Values>0.99 0.01 0.95 0.05 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="TuberculosisOrCancer"/>
          <Variable name="LungCancer"/>
          <Variable name="Tuberculosis"/>
        </Variables>
        <Values>1.0 0.0 0.0 1.0 0.0 1.0 0.0 1.0 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Sinus"/>
        </Variables>
        <Values>0.3 0.7 </Values>
      </Potential>
    </Potentials>
    <AdditionalConstraints/>
    <AdditionalProperties/>
    <DecisionCriteria/>
  </ProbNet>
</ProbModelXML>"""
        )
        self.assertDictEqual(model_data, model_data_expected)
        self.assertEqual(
            str(xmlfile.__str__()[:-1]), str(etree.tostring(xmlfile_expected))
        )



import unittest

import pandas as pd
import numpy as np

from pgmpy.estimators import MmhcEstimator, K2Score
from pgmpy.factors.discrete import TabularCPD
from pgmpy.models import BayesianModel


class TestMmhcEstimator(unittest.TestCase):
    def setUp(self):
        self.data1 = pd.DataFrame(
            np.random.randint(0, 2, size=(15000, 3)), columns=list("XYZ")
        )
        self.data1["sum"] = self.data1.sum(axis=1)
        self.est1 = MmhcEstimator(self.data1)

    @unittest.skip("currently disabled due to non-determenism")
    def test_estimate(self):
        self.assertTrue(
            set(self.est1.estimate().edges()).issubset(
                set(
                    [
                        ("X", "sum"),
                        ("Y", "sum"),
                        ("Z", "sum"),
                        ("sum", "X"),
                        ("sum", "Y"),
                        ("sum", "Z"),
                    ]
                )
            )
        )
        self.assertTrue(
            set(self.est1.estimate(significance_level=0.001).edges()).issubset(
                set(
                    [
                        ("X", "sum"),
                        ("Y", "sum"),
                        ("Z", "sum"),
                        ("sum", "X"),
                        ("sum", "Y"),
                        ("sum", "Z"),
                    ]
                )
            )
        )

    def tearDown(self):
        del self.data1
        del self.est1



from __future__ import division

from abc import ABCMeta, abstractmethod

import numpy as np
from scipy import integrate


class BaseDiscretizer(ABCMeta):
    """
    Base class for the discretizer classes in pgmpy. The discretizer
    classes are used to discretize a continuous random variable
    distribution into discrete probability masses.

    Parameters
    ----------
    factor: A ContinuousNode or a ContinuousFactor object
        the continuous node or factor representing the distribution
        to be discretized.

    low, high: float
        the range over which the function will be discretized.

    cardinality: int
        the number of states required in the discretized output.

    Examples
    --------
    >>> from scipy.stats import norm
    >>> from pgmpy.factors.continuous import ContinuousNode
    >>> normal = ContinuousNode(norm(0, 1).pdf)
    >>> from pgmpy.discretize import BaseDiscretizer
    >>> class ChildDiscretizer(BaseDiscretizer):
    ...     def get_discrete_values(self):
    ...         pass
    >>> discretizer = ChildDiscretizer(normal, -3, 3, 10)
    >>> discretizer.factor
    <pgmpy.factors.continuous.ContinuousNode.ContinuousNode object at 0x04C98190>
    >>> discretizer.cardinality
    10
    >>> discretizer.get_labels()
    ['x=-3.0', 'x=-2.4', 'x=-1.8', 'x=-1.2', 'x=-0.6', 'x=0.0', 'x=0.6', 'x=1.2', 'x=1.8', 'x=2.4']

    """

    def __init__(self, factor, low, high, cardinality):
        self.factor = factor
        self.low = low
        self.high = high
        self.cardinality = cardinality

    @abstractmethod
    def get_discrete_values(self):
        """
        This method implements the algorithm to discretize the given
        continuous distribution.

        It must be implemented by all the subclasses of BaseDiscretizer.

        Returns
        -------
        A list of discrete values or a DiscreteFactor object.
        """
        pass

    def get_labels(self):
        """
        Returns a list of strings representing the values about
        which the discretization method calculates the probabilty
        masses.

        Default value is the points -
        [low, low+step, low+2*step, ......... , high-step]
        unless the method is overridden by a subclass.

        Examples
        --------
        >>> from pgmpy.factors import ContinuousNode
        >>> from pgmpy.discretize import BaseDiscretizer
        >>> class ChildDiscretizer(BaseDiscretizer):
        ...     def get_discrete_values(self):
        ...         pass
        >>> from scipy.stats import norm
        >>> node = ContinuousNode(norm(0).pdf)
        >>> child = ChildDiscretizer(node, -5, 5, 20)
        >>> chld.get_labels()
        ['x=-5.0', 'x=-4.5', 'x=-4.0', 'x=-3.5', 'x=-3.0', 'x=-2.5',
         'x=-2.0', 'x=-1.5', 'x=-1.0', 'x=-0.5', 'x=0.0', 'x=0.5', 'x=1.0',
         'x=1.5', 'x=2.0', 'x=2.5', 'x=3.0', 'x=3.5', 'x=4.0', 'x=4.5']

        """
        step = (self.high - self.low) / self.cardinality
        labels = [
            "x={i}".format(i=str(i))
            for i in np.round(np.arange(self.low, self.high, step), 3)
        ]
        return labels


class RoundingDiscretizer(BaseDiscretizer):
    """
    This class uses the rounding method for discretizing the
    given continuous distribution.

    For the rounding method,

    The probability mass is,
    cdf(x+step/2)-cdf(x), for x = low

    cdf(x+step/2)-cdf(x-step/2), for low < x <= high

    where, cdf is the cumulative density function of the distribution
    and step = (high-low)/cardinality.

    Examples
    --------
    >>> import numpy as np
    >>> from pgmpy.factors.continuous import ContinuousNode
    >>> from pgmpy.factors.continuous import RoundingDiscretizer
    >>> std_normal_pdf = lambda x : np.exp(-x*x/2) / (np.sqrt(2*np.pi))
    >>> std_normal = ContinuousNode(std_normal_pdf)
    >>> std_normal.discretize(RoundingDiscretizer, low=-3, high=3,
    ...                       cardinality=12)
    [0.001629865203424451, 0.009244709419989363, 0.027834684208773178,
     0.065590616803038182, 0.120977578710013, 0.17466632194020804,
     0.19741265136584729, 0.17466632194020937, 0.12097757871001302,
     0.065590616803036905, 0.027834684208772664, 0.0092447094199902269]
    """

    def get_discrete_values(self):
        step = (self.high - self.low) / self.cardinality

        # for x=[low]
        discrete_values = [
            self.factor.cdf(self.low + step / 2) - self.factor.cdf(self.low)
        ]

        # for x=[low+step, low+2*step, ........., high-step]
        points = np.linspace(self.low + step, self.high - step, self.cardinality - 1)
        discrete_values.extend(
            [
                self.factor.cdf(i + step / 2) - self.factor.cdf(i - step / 2)
                for i in points
            ]
        )

        return discrete_values


class UnbiasedDiscretizer(BaseDiscretizer):
    """
    This class uses the unbiased method for discretizing the
    given continuous distribution.

    The unbiased method for discretization is the matching of the
    first moment method. It involves calculating the first order
    limited moment of the distribution which is done by the _lim_moment
    method.

    For this method,

    The probability mass is,
    (E(x) - E(x + step))/step + 1 - cdf(x), for x = low

    (2 * E(x) - E(x - step) - E(x + step))/step, for low < x < high

    (E(x) - E(x - step))/step - 1 + cdf(x), for x = high

    where, E(x) is the first limiting moment of the distribution
    about the point x, cdf is the cumulative density function
    and step = (high-low)/cardinality.

    References
    ----------
    Klugman, S. A., Panjer, H. H. and Willmot, G. E.,
    Loss Models, From Data to Decisions, Fourth Edition,
    Wiley, section 9.6.5.2 (Method of local monment matching) and
    exercise 9.41.

    Examples
    --------
    >>> import numpy as np
    >>> from pgmpy.factors import ContinuousNode
    >>> from pgmpy.factors.continuous import UnbiasedDiscretizer
    # exponential distribution with rate = 2
    >>> exp_pdf = lambda x: 2*np.exp(-2*x) if x>=0 else 0
    >>> exp_node = ContinuousNode(exp_pdf)
    >>> exp_node.discretize(UnbiasedDiscretizer, low=0, high=5, cardinality=10)
    [0.39627368905806137, 0.4049838434034298, 0.13331784003148325,
     0.043887287876647259, 0.014447413395300212, 0.0047559685431339703,
     0.0015656350182896128, 0.00051540201980112557, 0.00016965346326140994,
     3.7867260839208328e-05]

    """

    def get_discrete_values(self):
        lev = self._lim_moment
        step = (self.high - self.low) / (self.cardinality - 1)

        # for x=[low]
        discrete_values = [
            (lev(self.low) - lev(self.low + step)) / step
            + 1
            - self.factor.cdf(self.low)
        ]

        # for x=[low+step, low+2*step, ........., high-step]
        points = np.linspace(self.low + step, self.high - step, self.cardinality - 2)
        discrete_values.extend(
            [(2 * lev(i) - lev(i - step) - lev(i + step)) / step for i in points]
        )

        # for x=[high]
        discrete_values.append(
            (lev(self.high) - lev(self.high - step)) / step
            - 1
            + self.factor.cdf(self.high)
        )

        return discrete_values

    def _lim_moment(self, u, order=1):
        """
        This method calculates the kth order limiting moment of
        the distribution. It is given by -

        E(u) = Integral (-inf to u) [ (x^k)*pdf(x) dx ] + (u^k)(1-cdf(u))

        where, pdf is the probability density function and cdf is the
        cumulative density function of the distribution.

        Reference
        ---------
        Klugman, S. A., Panjer, H. H. and Willmot, G. E.,
        Loss Models, From Data to Decisions, Fourth Edition,
        Wiley, definition 3.5 and equation 3.8.

        Parameters
        ----------
        u: float
            The point at which the moment is to be calculated.

        order: int
            The order of the moment, default is first order.
        """

        def fun(x):
            return np.power(x, order) * self.factor.pdf(x)

        return integrate.quad(fun, -np.inf, u)[0] + np.power(u, order) * (
            1 - self.factor.cdf(u)
        )

    def get_labels(self):
        labels = list(
            "x={i}".format(i=str(i))
            for i in np.round(np.linspace(self.low, self.high, self.cardinality), 3)
        )
        return labels



#!/usr/bin/env python3
from collections import defaultdict
from warnings import warn

import numpy as np
from pandas import DataFrame
from scipy.linalg import eig

from pgmpy.factors.discrete import State
from pgmpy.utils import sample_discrete


class MarkovChain(object):
    """
    Class to represent a Markov Chain with multiple kernels for factored state space,
    along with methods to simulate a run.

    Examples
    --------

    Create an empty Markov Chain:

    >>> from pgmpy.models import MarkovChain as MC
    >>> model = MC()

    And then add variables to it

    >>> model.add_variables_from(['intel', 'diff'], [2, 3])

    Or directly create a Markov Chain from a list of variables and their cardinalities

    >>> model = MC(['intel', 'diff'], [2, 3])

    Add transition models

    >>> intel_tm = {0: {0: 0.25, 1: 0.75}, 1: {0: 0.5, 1: 0.5}}
    >>> model.add_transition_model('intel', intel_tm)
    >>> diff_tm = {0: {0: 0.1, 1: 0.5, 2: 0.4}, 1: {0: 0.2, 1: 0.2, 2: 0.6 }, 2: {0: 0.7, 1: 0.15, 2: 0.15}}
    >>> model.add_transition_model('diff', diff_tm)

    Set a start state

    >>> from pgmpy.factors.discrete import State
    >>> model.set_start_state([State('intel', 0), State('diff', 2)])

    Sample from it

    >>> model.sample(size=5)
       intel  diff
    0      0     2
    1      1     0
    2      0     1
    3      1     0
    4      0     2
    """

    def __init__(self, variables=None, card=None, start_state=None):
        """
        Parameters
        ----------
        variables: array-like iterable object
            A list of variables of the model.

        card: array-like iterable object
            A list of cardinalities of the variables.

        start_state: array-like iterable object
            List of tuples representing the starting states of the variables.
        """
        if variables is None:
            variables = []
        if card is None:
            card = []
        if not hasattr(variables, "__iter__") or isinstance(variables, str):
            raise ValueError("variables must be a non-string iterable.")
        if not hasattr(card, "__iter__") or isinstance(card, str):
            raise ValueError("card must be a non-string iterable.")
        self.variables = variables
        self.cardinalities = {v: c for v, c in zip(variables, card)}
        self.transition_models = {var: {} for var in variables}
        if start_state is None or self._check_state(start_state):
            self.state = start_state

    def set_start_state(self, start_state):
        """
        Set the start state of the Markov Chain. If the start_state is given as a array-like iterable, its contents
        are reordered in the internal representation.

        Parameters
        ----------
        start_state: dict or array-like iterable object
            Dict (or list) of tuples representing the starting states of the variables.

        Examples
        --------
        >>> from pgmpy.models import MarkovChain as MC
        >>> from pgmpy.factors.discrete import State
        >>> model = MC(['a', 'b'], [2, 2])
        >>> model.set_start_state([State('a', 0), State('b', 1)])
        """
        if start_state is not None:
            if not hasattr(start_state, "__iter__") or isinstance(start_state, str):
                raise ValueError("start_state must be a non-string iterable.")
            # Must be an array-like iterable. Reorder according to self.variables.
            state_dict = {var: st for var, st in start_state}
            start_state = [State(var, state_dict[var]) for var in self.variables]
        if start_state is None or self._check_state(start_state):
            self.state = start_state

    def _check_state(self, state):
        """
        Checks if a list representing the state of the variables is valid.
        """
        if not hasattr(state, "__iter__") or isinstance(state, str):
            raise ValueError("Start state must be a non-string iterable object.")
        state_vars = {s.var for s in state}
        if not state_vars == set(self.variables):
            raise ValueError(
                "Start state must represent a complete assignment to all variables."
                "Expected variables in state: {svar}, Got: {mvar}.".format(
                    svar=state_vars, mvar=set(self.variables)
                )
            )
        for var, val in state:
            if val >= self.cardinalities[var]:
                raise ValueError(
                    "Assignment {val} to {var} invalid.".format(val=val, var=var)
                )
        return True

    def add_variable(self, variable, card=0):
        """
        Add a variable to the model.

        Parameters
        ----------
        variable: any hashable python object

        card: int
            Representing the cardinality of the variable to be added.

        Examples
        --------
        >>> from pgmpy.models import MarkovChain as MC
        >>> model = MC()
        >>> model.add_variable('x', 4)
        """
        if variable not in self.variables:
            self.variables.append(variable)
        else:
            warn("Variable {var} already exists.".format(var=variable))
        self.cardinalities[variable] = card
        self.transition_models[variable] = {}

    def add_variables_from(self, variables, cards):
        """
        Add several variables to the model at once.

        Parameters
        ----------
        variables: array-like iterable object
            List of variables to be added.

        cards: array-like iterable object
            List of cardinalities of the variables to be added.

        Examples
        --------
        >>> from pgmpy.models import MarkovChain as MC
        >>> model = MC()
        >>> model.add_variables_from(['x', 'y'], [3, 4])
        """
        for var, card in zip(variables, cards):
            self.add_variable(var, card)

    def add_transition_model(self, variable, transition_model):
        """
        Adds a transition model for a particular variable.

        Parameters
        ----------
        variable: any hashable python object
            must be an existing variable of the model.
        transition_model: dict or 2d array
            dict representing valid transition probabilities defined for every possible state of the variable.
            array represent a square matrix where every row sums to 1,
            array[i,j] indicates the transition probalities from State i to State j


        Examples
        --------
        >>> from pgmpy.models import MarkovChain as MC
        >>> model = MC()
        >>> model.add_variable('grade', 3)
        >>> grade_tm = {0: {0: 0.1, 1: 0.5, 2: 0.4}, 1: {0: 0.2, 1: 0.2, 2: 0.6 }, 2: {0: 0.7, 1: 0.15, 2: 0.15}}
        >>> grade_tm_matrix = np.array([[0.1, 0.5, 0.4], [0.2, 0.2, 0.6], [0.7, 0.15, 0.15]])
        >>> model.add_transition_model('grade', grade_tm)
        >>> model.add_transition_model('grade', grade_tm_matrix)
        """
        if isinstance(transition_model, list):
            transition_model = np.array(transition_model)

        # check if the transition model is valid
        if not isinstance(transition_model, dict):
            if not isinstance(transition_model, np.ndarray):
                raise ValueError("Transition model must be a dict or numpy array")
            elif len(transition_model.shape) != 2:
                raise ValueError(
                    "Transition model must be 2d array.given {t}".format(
                        t=transition_model.shape
                    )
                )
            elif transition_model.shape[0] != transition_model.shape[1]:
                raise ValueError(
                    "Dimension mismatch {d1}!={d2}".format(
                        d1=transition_model.shape[0], d2=transition_model.shape[1]
                    )
                )
            else:
                # convert the matrix to dict
                size = transition_model.shape[0]
                transition_model = dict(
                    (
                        i,
                        dict(
                            (j, float(transition_model[i][j])) for j in range(0, size)
                        ),
                    )
                    for i in range(0, size)
                )

        exp_states = set(range(self.cardinalities[variable]))
        tm_states = set(transition_model.keys())
        if not exp_states == tm_states:
            raise ValueError(
                "Transitions must be defined for all states of variable {v}. "
                "Expected states: {es}, Got: {ts}.".format(
                    v=variable, es=exp_states, ts=tm_states
                )
            )

        for _, transition in transition_model.items():
            if not isinstance(transition, dict):
                raise ValueError("Each transition must be a dict.")
            prob_sum = 0

            for _, prob in transition.items():
                if prob < 0 or prob > 1:
                    raise ValueError(
                        "Transitions must represent valid probability weights."
                    )
                prob_sum += prob

            if not np.allclose(prob_sum, 1):
                raise ValueError("Transition probabilities must sum to 1.")

        self.transition_models[variable] = transition_model

    def sample(self, start_state=None, size=1):
        """
        Sample from the Markov Chain.

        Parameters
        ----------
        start_state: dict or array-like iterable
            Representing the starting states of the variables. If None is passed, a random start_state is chosen.
        size: int
            Number of samples to be generated.

        Returns
        -------
        pandas.DataFrame

        Examples
        --------
        >>> from pgmpy.models import MarkovChain as MC
        >>> from pgmpy.factors.discrete import State
        >>> model = MC(['intel', 'diff'], [2, 3])
        >>> model.set_start_state([State('intel', 0), State('diff', 2)])
        >>> intel_tm = {0: {0: 0.25, 1: 0.75}, 1: {0: 0.5, 1: 0.5}}
        >>> model.add_transition_model('intel', intel_tm)
        >>> diff_tm = {0: {0: 0.1, 1: 0.5, 2: 0.4}, 1: {0: 0.2, 1: 0.2, 2: 0.6 }, 2: {0: 0.7, 1: 0.15, 2: 0.15}}
        >>> model.add_transition_model('diff', diff_tm)
        >>> model.sample(size=5)
           intel  diff
        0      0     2
        1      1     0
        2      0     1
        3      1     0
        4      0     2
        """
        if start_state is None:
            if self.state is None:
                self.state = self.random_state()
            # else use previously-set state
        else:
            self.set_start_state(start_state)

        sampled = DataFrame(index=range(size), columns=self.variables)
        sampled.loc[0] = [st for var, st in self.state]

        var_states = defaultdict(dict)
        var_values = defaultdict(dict)
        samples = defaultdict(dict)
        for var in self.transition_models.keys():
            for st in self.transition_models[var]:
                var_states[var][st] = list(self.transition_models[var][st].keys())
                var_values[var][st] = list(self.transition_models[var][st].values())
                samples[var][st] = sample_discrete(
                    var_states[var][st], var_values[var][st], size=size
                )

        for i in range(size - 1):
            for j, (var, st) in enumerate(self.state):
                next_st = samples[var][st][i]
                self.state[j] = State(var, next_st)
            sampled.loc[i + 1] = [st for var, st in self.state]

        return sampled

    def prob_from_sample(self, state, sample=None, window_size=None):
        """
        Given an instantiation (partial or complete) of the variables of the model,
        compute the probability of observing it over multiple windows in a given sample.

        If 'sample' is not passed as an argument, generate the statistic by sampling from the
        Markov Chain, starting with a random initial state.

        Examples
        --------
        >>> from pgmpy.models.MarkovChain import MarkovChain as MC
        >>> from pgmpy.factors.discrete import State
        >>> model = MC(['intel', 'diff'], [3, 2])
        >>> intel_tm = {0: {0: 0.2, 1: 0.4, 2:0.4}, 1: {0: 0, 1: 0.5, 2: 0.5}, 2: {2: 0.5, 1:0.5}}
        >>> model.add_transition_model('intel', intel_tm)
        >>> diff_tm = {0: {0: 0.5, 1: 0.5}, 1: {0: 0.25, 1:0.75}}
        >>> model.add_transition_model('diff', diff_tm)
        >>> model.prob_from_sample([State('diff', 0)])
        array([ 0.27,  0.4 ,  0.18,  0.23, ..., 0.29])
        """
        if sample is None:
            # generate sample of size 10000
            sample = self.sample(self.random_state(), size=10000)
        if window_size is None:
            window_size = len(sample) // 100  # default window size is 100
        windows = len(sample) // window_size
        probabilities = np.zeros(windows)

        for i in range(windows):
            for j in range(window_size):
                ind = i * window_size + j
                state_eq = [sample.loc[ind, v] == s for v, s in state]
                if all(state_eq):
                    probabilities[i] += 1

        return probabilities / window_size

    def generate_sample(self, start_state=None, size=1):
        """
        Generator version of self.sample

        Returns
        -------
        List of State namedtuples, representing the assignment to all variables of the model.

        Examples
        --------
        >>> from pgmpy.models.MarkovChain import MarkovChain
        >>> from pgmpy.factors.discrete import State
        >>> model = MarkovChain()
        >>> model.add_variables_from(['intel', 'diff'], [3, 2])
        >>> intel_tm = {0: {0: 0.2, 1: 0.4, 2:0.4}, 1: {0: 0, 1: 0.5, 2: 0.5}, 2: {0: 0.3, 1: 0.3, 2: 0.4}}
        >>> model.add_transition_model('intel', intel_tm)
        >>> diff_tm = {0: {0: 0.5, 1: 0.5}, 1: {0: 0.25, 1:0.75}}
        >>> model.add_transition_model('diff', diff_tm)
        >>> gen = model.generate_sample([State('intel', 0), State('diff', 0)], 2)
        >>> [sample for sample in gen]
        [[State(var='intel', state=2), State(var='diff', state=1)],
         [State(var='intel', state=2), State(var='diff', state=0)]]
        """
        if start_state is None:
            if self.state is None:
                self.state = self.random_state()
            # else use previously-set state
        else:
            self.set_start_state(start_state)
        # sampled.loc[0] = [self.state[var] for var in self.variables]

        for i in range(size):
            for j, (var, st) in enumerate(self.state):
                next_st = sample_discrete(
                    list(self.transition_models[var][st].keys()),
                    list(self.transition_models[var][st].values()),
                )[0]
                self.state[j] = State(var, next_st)
            yield self.state[:]

    def is_stationarity(self, tolerance=0.2, sample=None):
        """
        Checks if the given markov chain is stationary and checks the steady state
        probablity values for the state are consistent.

        Parameters
        ----------
        tolerance: float
            represents the diff between actual steady state value and the computed value
        sample: [State(i,j)]
            represents the list of state which the markov chain has sampled

        Returns
        -------
        Boolean:
            True, if the markov chain converges to steady state distribution within the tolerance
            False, if the markov chain does not converge to steady state distribution within tolerance

        Examples
        --------
        >>> from pgmpy.models.MarkovChain import MarkovChain
        >>> from pgmpy.factors.discrete import State
        >>> model = MarkovChain()
        >>> model.add_variables_from(['intel', 'diff'], [3, 2])
        >>> intel_tm = {0: {0: 0.2, 1: 0.4, 2:0.4}, 1: {0: 0, 1: 0.5, 2: 0.5}, 2: {0: 0.3, 1: 0.3, 2: 0.4}}
        >>> model.add_transition_model('intel', intel_tm)
        >>> diff_tm = {0: {0: 0.5, 1: 0.5}, 1: {0: 0.25, 1:0.75}}
        >>> model.add_transition_model('diff', diff_tm)
        >>> model.is_stationarity()
        True
        """
        keys = self.transition_models.keys()
        return_val = True
        for k in keys:
            # convert dict to numpy matrix
            transition_mat = np.array(
                [
                    np.array(list(self.transition_models[k][i].values()))
                    for i in self.transition_models[k].keys()
                ],
                dtype=np.float,
            )
            S, U = eig(transition_mat.T)
            stationary = np.array(U[:, np.where(np.abs(S - 1.0) < 1e-8)[0][0]].flat)
            stationary = (stationary / np.sum(stationary)).real

            probabilites = []
            window_size = 10000 if sample is None else len(sample)
            for i in range(0, transition_mat.shape[0]):
                probabilites.extend(
                    self.prob_from_sample([State(k, i)], window_size=window_size)
                )
            if any(
                np.abs(i) > tolerance for i in np.subtract(probabilites, stationary)
            ):
                return_val = return_val and False
            else:
                return_val = return_val and True

        return return_val

    def random_state(self):
        """
        Generates a random state of the Markov Chain.

        Returns
        -------
        List of namedtuples, representing a random assignment to all variables of the model.

        Examples
        --------
        >>> from pgmpy.models import MarkovChain as MC
        >>> model = MC(['intel', 'diff'], [2, 3])
        >>> model.random_state()
        [State('diff', 2), State('intel', 1)]
        """
        return [
            State(var, np.random.randint(self.cardinalities[var]))
            for var in self.variables
        ]

    def copy(self):
        """
        Returns a copy of Markov Chain Model.

        Returns
        -------
        MarkovChain : Copy of MarkovChain.

        Examples
        --------
        >>> from pgmpy.models import MarkovChain
        >>> from pgmpy.factors.discrete import State
        >>> model = MarkovChain()
        >>> model.add_variables_from(['intel', 'diff'], [3, 2])
        >>> intel_tm = {0: {0: 0.2, 1: 0.4, 2:0.4}, 1: {0: 0, 1: 0.5, 2: 0.5}, 2: {0: 0.3, 1: 0.3, 2: 0.4}}
        >>> model.add_transition_model('intel', intel_tm)
        >>> diff_tm = {0: {0: 0.5, 1: 0.5}, 1: {0: 0.25, 1:0.75}}
        >>> model.add_transition_model('diff', diff_tm)
        >>> model.set_start_state([State('intel', 0), State('diff', 2)])
        >>> model_copy = model.copy()
        >>> model_copy.transition_models
        >>> {'diff': {0: {0: 0.1, 1: 0.5, 2: 0.4}, 1: {0: 0.2, 1: 0.2, 2: 0.6}, 2: {0: 0.7, 1: 0.15, 2: 0.15}},
             'intel': {0: {0: 0.25, 1: 0.75}, 1: {0: 0.5, 1: 0.5}}}
        """
        markovchain_copy = MarkovChain(
            variables=list(self.cardinalities.keys()),
            card=list(self.cardinalities.values()),
            start_state=self.state,
        )
        if self.transition_models:
            markovchain_copy.transition_models = self.transition_models.copy()

        return markovchain_copy



from __future__ import division

import numpy as np
import networkx as nx

from pgmpy.models import BayesianModel
from pgmpy.factors.continuous import LinearGaussianCPD
from pgmpy.factors.distributions import GaussianDistribution


class LinearGaussianBayesianNetwork(BayesianModel):
    """
    A Linear Gaussain Bayesian Network is a Bayesian Network, all
    of whose variables are continuous, and where all of the CPDs
    are linear Gaussians.

    An important result is that the linear Gaussian Bayesian Networks
    are an alternative representation for the class of multivariate
    Gaussian distributions.

    """

    def add_cpds(self, *cpds):
        """
        Add linear Gaussian CPD (Conditional Probability Distribution)
        to the Bayesian Model.

        Parameters
        ----------
        cpds  :  instances of LinearGaussianCPD
            List of LinearGaussianCPDs which will be associated with the model

        Examples
        --------
        >>> from pgmpy.models import LinearGaussianBayesianNetwork
        >>> from pgmpy.factors.continuous import LinearGaussianCPD
        >>> model = LinearGaussianBayesianNetwork([('x1', 'x2'), ('x2', 'x3')])
        >>> cpd1 = LinearGaussianCPD('x1', [1], 4)
        >>> cpd2 = LinearGaussianCPD('x2', [-5, 0.5], 4, ['x1'])
        >>> cpd3 = LinearGaussianCPD('x3', [4, -1], 3, ['x2'])
        >>> model.add_cpds(cpd1, cpd2, cpd3)
        >>> for cpd in model.cpds:
        ...     print(cpd)

        P(x1) = N(1; 4)
        P(x2| x1) = N(0.5*x1_mu); -5)
        P(x3| x2) = N(-1*x2_mu); 4)

        """
        for cpd in cpds:
            if not isinstance(cpd, LinearGaussianCPD):
                raise ValueError("Only LinearGaussianCPD can be added.")

            if set(cpd.variables) - set(cpd.variables).intersection(set(self.nodes())):
                raise ValueError("CPD defined on variable not in the model", cpd)

            for prev_cpd_index in range(len(self.cpds)):
                if self.cpds[prev_cpd_index].variable == cpd.variable:
                    logging.warning(
                        "Replacing existing CPD for {var}".format(var=cpd.variable)
                    )
                    self.cpds[prev_cpd_index] = cpd
                    break
            else:
                self.cpds.append(cpd)

    def get_cpds(self, node=None):
        """
        Returns the cpd of the node. If node is not specified returns all the CPDs
        that have been added till now to the graph

        Parameter
        ---------
        node: any hashable python object (optional)
            The node whose CPD we want. If node not specified returns all the
            CPDs added to the model.

        Returns
        -------
        A list of linear Gaussian CPDs.

        Examples
        --------
        >>> from pgmpy.models import LinearGaussianBayesianNetwork
        >>> from pgmpy.factors.continuous import LinearGaussianCPD
        >>> model = LinearGaussianBayesianNetwork([('x1', 'x2'), ('x2', 'x3')])
        >>> cpd1 = LinearGaussianCPD('x1', [1], 4)
        >>> cpd2 = LinearGaussianCPD('x2', [-5, 0.5], 4, ['x1'])
        >>> cpd3 = LinearGaussianCPD('x3', [4, -1], 3, ['x2'])
        >>> model.add_cpds(cpd1, cpd2, cpd3)
        >>> model.get_cpds()
        """
        return super(LinearGaussianBayesianNetwork, self).get_cpds(node)

    def remove_cpds(self, *cpds):
        """
        Removes the cpds that are provided in the argument.

        Parameters
        ----------
        *cpds: LinearGaussianCPD object
            A LinearGaussianCPD object on any subset of the variables
            of the model which is to be associated with the model.

        Examples
        --------
        >>> from pgmpy.models import LinearGaussianBayesianNetwork
        >>> from pgmpy.factors.continuous import LinearGaussianCPD
        >>> model = LinearGaussianBayesianNetwork([('x1', 'x2'), ('x2', 'x3')])
        >>> cpd1 = LinearGaussianCPD('x1', [1], 4)
        >>> cpd2 = LinearGaussianCPD('x2', [-5, 0.5], 4, ['x1'])
        >>> cpd3 = LinearGaussianCPD('x3', [4, -1], 3, ['x2'])
        >>> model.add_cpds(cpd1, cpd2, cpd3)
        >>> for cpd in model.get_cpds():
        ...     print(cpd)

        P(x1) = N(1; 4)
        P(x2| x1) = N(0.5*x1_mu); -5)
        P(x3| x2) = N(-1*x2_mu); 4)

        >>> model.remove_cpds(cpd2, cpd3)
        >>> for cpd in model.get_cpds():
        ...     print(cpd)

        P(x1) = N(1; 4)

        """
        return super(LinearGaussianBayesianNetwork, self).remove_cpds(*cpds)

    def to_joint_gaussian(self):
        """
        The linear Gaussian Bayesian Networks are an alternative
        representation for the class of multivariate Gaussian distributions.
        This method returns an equivalent joint Gaussian distribution.

        Returns
        -------
        GaussianDistribution: An equivalent joint Gaussian
                                   distribution for the network.

        Reference
        ---------
        Section 7.2, Example 7.3,
        Probabilistic Graphical Models, Principles and Techniques

        Examples
        --------
        >>> from pgmpy.models import LinearGaussianBayesianNetwork
        >>> from pgmpy.factors.continuous import LinearGaussianCPD
        >>> model = LinearGaussianBayesianNetwork([('x1', 'x2'), ('x2', 'x3')])
        >>> cpd1 = LinearGaussianCPD('x1', [1], 4)
        >>> cpd2 = LinearGaussianCPD('x2', [-5, 0.5], 4, ['x1'])
        >>> cpd3 = LinearGaussianCPD('x3', [4, -1], 3, ['x2'])
        >>> model.add_cpds(cpd1, cpd2, cpd3)
        >>> jgd = model.to_joint_gaussian()
        >>> jgd.variables
        ['x1', 'x2', 'x3']
        >>> jgd.mean
        array([[ 1. ],
               [-4.5],
               [ 8.5]])
        >>> jgd.covariance
        array([[ 4.,  2., -2.],
               [ 2.,  5., -5.],
               [-2., -5.,  8.]])

        """
        variables = list(nx.topological_sort(self))
        mean = np.zeros(len(variables))
        covariance = np.zeros((len(variables), len(variables)))

        for node_idx in range(len(variables)):
            cpd = self.get_cpds(variables[node_idx])
            mean[node_idx] = (
                sum(
                    [
                        coeff * mean[variables.index(parent)]
                        for coeff, parent in zip(cpd.mean, cpd.evidence)
                    ]
                )
                + cpd.mean[0]
            )
            covariance[node_idx, node_idx] = (
                sum(
                    [
                        coeff
                        * coeff
                        * covariance[variables.index(parent), variables.index(parent)]
                        for coeff, parent in zip(cpd.mean, cpd.evidence)
                    ]
                )
                + cpd.variance
            )

        for node_i_idx in range(len(variables)):
            for node_j_idx in range(len(variables)):
                if covariance[node_j_idx, node_i_idx] != 0:
                    covariance[node_i_idx, node_j_idx] = covariance[
                        node_j_idx, node_i_idx
                    ]
                else:
                    cpd_j = self.get_cpds(variables[node_j_idx])
                    covariance[node_i_idx, node_j_idx] = sum(
                        [
                            coeff * covariance[node_i_idx, variables.index(parent)]
                            for coeff, parent in zip(cpd_j.mean, cpd_j.evidence)
                        ]
                    )

        return GaussianDistribution(variables, mean, covariance)

    def check_model(self):
        """
        Checks the model for various errors. This method checks for the following
        error -

        * Checks if the CPDs associated with nodes are consistent with their parents.

        Returns
        -------
        check: boolean
            True if all the checks pass.

        """
        for node in self.nodes():
            cpd = self.get_cpds(node=node)

            if isinstance(cpd, LinearGaussianCPD):
                if set(cpd.evidence) != set(self.get_parents(node)):
                    raise ValueError(
                        "CPD associated with %s doesn't have "
                        "proper parents associated with it." % node
                    )
        return True

    def get_cardinality(self, node):
        """
        Cardinality is not defined for continuous variables.
        """
        raise ValueError("Cardinality is not defined for continuous variables.")

    def fit(
        self, data, estimator=None, state_names=[], complete_samples_only=True, **kwargs
    ):
        """
        For now, fit method has not been implemented for LinearGaussianBayesianNetwork.
        """

        raise NotImplementedError(
            "fit method has not been implemented for LinearGaussianBayesianNetwork."
        )

    def predict(self, data):
        """
        For now, predict method has not been implemented for LinearGaussianBayesianNetwork.
        """
        raise NotImplementedError(
            "predict method has not been implemented for LinearGaussianBayesianNetwork."
        )

    def to_markov_model(self):
        """
        For now, to_markov_model method has not been implemented for LinearGaussianBayesianNetwork.
        """
        raise NotImplementedError(
            "to_markov_model method has not been implemented for LinearGaussianBayesianNetwork."
        )

    def is_imap(self, JPD):
        """
        For now, is_imap method has not been implemented for LinearGaussianBayesianNetwork.
        """
        raise NotImplementedError(
            "is_imap method has not been implemented for LinearGaussianBayesianNetwork."
        )



import os
import unittest

import numpy as np
import numpy.testing as np_test
import networkx as nx

from pgmpy.readwrite import BIFReader, BIFWriter
from pgmpy.models import BayesianModel
from pgmpy.factors.discrete import TabularCPD


class TestBIFReader(unittest.TestCase):
    def setUp(self):
        self.reader = BIFReader(
            string="""
                // Bayesian Network in the Interchange Format
                // Produced by BayesianNetworks package in JavaBayes
                // Output created Sun Nov 02 17:49:49 GMT+00:00 1997
                // Bayesian network
                network "Dog-Problem" { //5 variables and 5 probability distributions
                        property "credal-set constant-density-bounded 1.1" ;
                }
                variable  "light-on" { //2 values
                        type discrete[2] {  "true"  "false" };
                        property "position = (218, 195)" ;
                }
                variable  "bowel-problem" { //2 values
                        type discrete[2] {  "true"  "false" };
                        property "position = (335, 99)" ;
                }
                variable  "dog-out" { //2 values
                        type discrete[2] {  "true"  "false" };
                        property "position = (300, 195)" ;
                }
                variable  "hear-bark" { //2 values
                        type discrete[2] {  "true"  "false" };
                        property "position = (296, 268)" ;
                }
                variable  "family-out" { //2 values
                        type discrete[2] {  "true"  "false" };
                        property "position = (257, 99)" ;
                }
                probability (  "light-on"  "family-out" ) { //2 variable(s) and 4 values
                        (true) 0.6 0.4 ;
                        (false) 0.05 0.95 ;
                }
                probability (  "bowel-problem" ) { //1 variable(s) and 2 values
                        table 0.01 0.99 ;
                }
                probability (  "dog-out"  "bowel-problem"  "family-out" ) { //3 variable(s) and 8 values
                        table 0.99 0.97 0.9 0.3 0.01 0.03 0.1 0.7 ;
                }
                probability (  "hear-bark"  "dog-out" ) { //2 variable(s) and 4 values
                        table 0.7 0.01 0.3 0.99 ;
                }
                probability (  "family-out" ) { //1 variable(s) and 2 values
                        table 0.15 0.85 ;
                }
                """,
            include_properties=True,
        )

        self.water_model = BIFReader(
            "pgmpy/tests/test_readwrite/testdata/water.bif", include_properties=True
        )

    def test_network_name(self):

        name_expected = "Dog-Problem"
        self.assertEqual(self.reader.network_name, name_expected)

    def test_get_variables(self):

        var_expected = [
            "light-on",
            "bowel-problem",
            "dog-out",
            "hear-bark",
            "family-out",
        ]
        self.assertListEqual(self.reader.get_variables(), var_expected)

    def test_states(self):

        states_expected = {
            "bowel-problem": ["true", "false"],
            "dog-out": ["true", "false"],
            "family-out": ["true", "false"],
            "hear-bark": ["true", "false"],
            "light-on": ["true", "false"],
        }
        states = self.reader.get_states()
        for variable in states_expected:
            self.assertListEqual(states_expected[variable], states[variable])

    def test_get_property(self):

        property_expected = {
            "bowel-problem": ["position = (335, 99)"],
            "dog-out": ["position = (300, 195)"],
            "family-out": ["position = (257, 99)"],
            "hear-bark": ["position = (296, 268)"],
            "light-on": ["position = (218, 195)"],
        }
        prop = self.reader.get_property()
        for variable in property_expected:
            self.assertListEqual(property_expected[variable], prop[variable])

    def test_get_values(self):

        cpd_expected = {
            "bowel-problem": np.array([[0.01], [0.99]]),
            "dog-out": np.array([[0.99, 0.97, 0.9, 0.3], [0.01, 0.03, 0.1, 0.7]]),
            "family-out": np.array([[0.15], [0.85]]),
            "hear-bark": np.array([[0.7, 0.01], [0.3, 0.99]]),
            "light-on": np.array([[0.6, 0.05], [0.4, 0.95]]),
        }
        cpd = self.reader.variable_cpds
        for variable in cpd_expected:
            np_test.assert_array_equal(cpd_expected[variable], cpd[variable])

    def test_get_values_reordered(self):

        cancer_values1 = BIFReader(
            string="""
                network unknown {
                }
                variable Pollution {
                  type discrete [ 2 ] { low, high };
                }
                variable Smoker {
                  type discrete [ 2 ] { True, False };
                }
                variable Cancer {
                  type discrete [ 2 ] { True, False };
                }
                probability ( Cancer | Pollution, Smoker ) {
                  (low, True) 0.03, 0.97;
                  (low, False) 0.001, 0.999;
                  (high, True) 0.05, 0.95;
                  (high, False) 0.02, 0.98;
                }"""
        ).get_values()

        cancer_values2 = BIFReader(
            string="""
                network unknown {
                }
                variable Pollution {
                  type discrete [ 2 ] { low, high };
                }
                variable Smoker {
                  type discrete [ 2 ] { True, False };
                }
                variable Cancer {
                  type discrete [ 2 ] { True, False };
                }
                probability ( Cancer | Pollution, Smoker ) {
                  (low, True) 0.03, 0.97;
                  (high, True) 0.05, 0.95;
                  (low, False) 0.001, 0.999;
                  (high, False) 0.02, 0.98;
                }"""
        ).get_values()

        for var in cancer_values1:
            np_test.assert_array_equal(cancer_values1[var], cancer_values2[var])

    def test_get_parents(self):

        parents_expected = {
            "bowel-problem": [],
            "dog-out": ["bowel-problem", "family-out"],
            "family-out": [],
            "hear-bark": ["dog-out"],
            "light-on": ["family-out"],
        }
        parents = self.reader.get_parents()
        for variable in parents_expected:
            self.assertListEqual(parents_expected[variable], parents[variable])

    def test_get_edges(self):

        edges_expected = [
            ["family-out", "dog-out"],
            ["bowel-problem", "dog-out"],
            ["family-out", "light-on"],
            ["dog-out", "hear-bark"],
        ]
        self.assertListEqual(sorted(self.reader.variable_edges), sorted(edges_expected))

    def test_get_model(self):
        edges_expected = [
            ("family-out", "dog-out"),
            ("bowel-problem", "dog-out"),
            ("family-out", "light-on"),
            ("dog-out", "hear-bark"),
        ]
        nodes_expected = [
            "bowel-problem",
            "hear-bark",
            "light-on",
            "dog-out",
            "family-out",
        ]
        edge_expected = {
            "bowel-problem": {"dog-out": {"weight": None}},
            "dog-out": {"hear-bark": {"weight": None}},
            "family-out": {"dog-out": {"weight": None}, "light-on": {"weight": None}},
            "hear-bark": {},
            "light-on": {},
        }
        node_expected = {
            "bowel-problem": {"weight": None, "position": "(335, 99)"},
            "dog-out": {"weight": None, "position": "(300, 195)"},
            "family-out": {"weight": None, "position": "(257, 99)"},
            "hear-bark": {"weight": None, "position": "(296, 268)"},
            "light-on": {"weight": None, "position": "(218, 195)"},
        }
        cpds_expected = [
            TabularCPD(
                variable="bowel-problem",
                variable_card=2,
                values=np.array([[0.01], [0.99]]),
                state_names={"bowel-problem": ["true", "false"]},
            ),
            TabularCPD(
                variable="dog-out",
                variable_card=2,
                values=np.array([[0.99, 0.97, 0.9, 0.3], [0.01, 0.03, 0.1, 0.7]]),
                evidence=["bowel-problem", "family-out"],
                evidence_card=[2, 2],
                state_names={
                    "dog-out": ["true", "false"],
                    "bowel-problem": ["true", "false"],
                    "family-out": ["true", "false"],
                },
            ),
            TabularCPD(
                variable="family-out",
                variable_card=2,
                values=np.array([[0.15], [0.85]]),
                state_names={"family-out": ["true", "false"]},
            ),
            TabularCPD(
                variable="hear-bark",
                variable_card=2,
                values=np.array([[0.7, 0.01], [0.3, 0.99]]),
                evidence=["dog-out"],
                evidence_card=[2],
                state_names={
                    "hear-bark": ["true", "false"],
                    "dog-out": ["true", "false"],
                },
            ),
            TabularCPD(
                variable="light-on",
                variable_card=2,
                values=np.array([[0.6, 0.05], [0.4, 0.95]]),
                evidence=["family-out"],
                evidence_card=[2],
                state_names={
                    "light-on": ["true", "false"],
                    "family-out": ["true", "false"],
                },
            ),
        ]
        model = self.reader.get_model()
        model_cpds = model.get_cpds()
        for cpd_index in range(5):
            self.assertEqual(model_cpds[cpd_index], cpds_expected[cpd_index])

        self.assertDictEqual(dict(model.nodes), node_expected)
        self.assertDictEqual(dict(model.adj), edge_expected)

        self.assertListEqual(sorted(model.nodes()), sorted(nodes_expected))
        self.assertListEqual(sorted(model.edges()), sorted(edges_expected))

    def test_water_model(self):
        model = self.water_model.get_model()
        self.assertEqual(len(model.nodes()), 32)
        self.assertEqual(len(model.edges()), 66)
        self.assertEqual(len(model.get_cpds()), 32)

    def tearDown(self):
        del self.reader


class TestBIFWriter(unittest.TestCase):
    def setUp(self):
        variables = [
            "kid",
            "bowel-problem",
            "dog-out",
            "family-out",
            "hear-bark",
            "light-on",
        ]

        edges = [
            ["family-out", "dog-out"],
            ["bowel-problem", "dog-out"],
            ["family-out", "light-on"],
            ["dog-out", "hear-bark"],
        ]

        cpds = {
            "kid": np.array([[0.3], [0.7]]),
            "bowel-problem": np.array([[0.01], [0.99]]),
            "dog-out": np.array([[0.99, 0.01, 0.97, 0.03], [0.9, 0.1, 0.3, 0.7]]),
            "family-out": np.array([[0.15], [0.85]]),
            "hear-bark": np.array([[0.7, 0.3], [0.01, 0.99]]),
            "light-on": np.array([[0.6, 0.4], [0.05, 0.95]]),
        }

        states = {
            "kid": ["true", "false"],
            "bowel-problem": ["true", "false"],
            "dog-out": ["true", "false"],
            "family-out": ["true", "false"],
            "hear-bark": ["true", "false"],
            "light-on": ["true", "false"],
        }

        parents = {
            "kid": [],
            "bowel-problem": [],
            "dog-out": ["bowel-problem", "family-out"],
            "family-out": [],
            "hear-bark": ["dog-out"],
            "light-on": ["family-out"],
        }

        properties = {
            "kid": ["position = (100, 165)"],
            "bowel-problem": ["position = (335, 99)"],
            "dog-out": ["position = (300, 195)"],
            "family-out": ["position = (257, 99)"],
            "hear-bark": ["position = (296, 268)"],
            "light-on": ["position = (218, 195)"],
        }

        self.model = BayesianModel()
        self.model.add_nodes_from(variables)
        self.model.add_edges_from(edges)

        tabular_cpds = []
        for var in sorted(cpds.keys()):
            values = cpds[var]
            cpd = TabularCPD(
                var,
                len(states[var]),
                values,
                evidence=parents[var],
                evidence_card=[
                    len(states[evidence_var]) for evidence_var in parents[var]
                ],
            )
            tabular_cpds.append(cpd)
        self.model.add_cpds(*tabular_cpds)

        for node, properties in properties.items():
            for prop in properties:
                prop_name, prop_value = map(lambda t: t.strip(), prop.split("="))
                self.model.nodes[node][prop_name] = prop_value

        self.writer = BIFWriter(model=self.model)

    def test_str(self):
        self.expected_string = """network unknown {
}
variable bowel-problem {
    type discrete [ 2 ] { 0, 1 };
    property position = (335, 99) ;
    property weight = None ;
}
variable dog-out {
    type discrete [ 2 ] { 0, 1 };
    property position = (300, 195) ;
    property weight = None ;
}
variable family-out {
    type discrete [ 2 ] { 0, 1 };
    property position = (257, 99) ;
    property weight = None ;
}
variable hear-bark {
    type discrete [ 2 ] { 0, 1 };
    property position = (296, 268) ;
    property weight = None ;
}
variable kid {
    type discrete [ 2 ] { 0, 1 };
    property position = (100, 165) ;
    property weight = None ;
}
variable light-on {
    type discrete [ 2 ] { 0, 1 };
    property position = (218, 195) ;
    property weight = None ;
}
probability ( bowel-problem ) {
    table 0.01, 0.99 ;
}
probability ( dog-out | bowel-problem, family-out ) {
    table 0.99, 0.01, 0.97, 0.03, 0.9, 0.1, 0.3, 0.7 ;
}
probability ( family-out ) {
    table 0.15, 0.85 ;
}
probability ( hear-bark | dog-out ) {
    table 0.7, 0.3, 0.01, 0.99 ;
}
probability ( kid ) {
    table 0.3, 0.7 ;
}
probability ( light-on | family-out ) {
    table 0.6, 0.4, 0.05, 0.95 ;
}
"""
        self.maxDiff = None
        self.assertEqual(self.writer.__str__(), self.expected_string)

    @unittest.skip("Fix this when #1221 is resolved")
    def test_write_read_equal(self):
        self.writer.write_bif("test_bif.bif")
        reader = BIFReader("test_bif.bif")
        read_model = reader.get_model()
        self.assertEqual(sorted(self.model.nodes()), sorted(read_model.nodes()))
        self.assertEqual(sorted(self.model.edges()), sorted(read_model.edges()))
        for var in self.model.nodes():
            self.assertEqual(self.model.get_cpds(var), read_model.get_cpds(var))
        os.remove("test_bif.bif")



from warnings import warn

import numpy as np
import pandas as pd

from scipy import stats


def chi_square(X, Y, Z, data, **kwargs):
    """
    Chi-square conditional independence test.
    Tests the null hypothesis that X is independent from Y given Zs.

    This is done by comparing the observed frequencies with the expected
    frequencies if X,Y were conditionally independent, using a chisquare
    deviance statistic. The expected frequencies given independence are
    `P(X,Y,Zs) = P(X|Zs)*P(Y|Zs)*P(Zs)`. The latter term can be computed
    as `P(X,Zs)*P(Y,Zs)/P(Zs).

    Parameters
    ----------
    X: int, string, hashable object
        A variable name contained in the data set
    Y: int, string, hashable object
        A variable name contained in the data set, different from X
    Zs: list of variable names
        A list of variable names contained in the data set, different from X and Y.
        This is the separating set that (potentially) makes X and Y independent.
        Default: []

    Returns
    -------
    chi2: float
        The chi2 test statistic.
    p_value: float
        The p_value, i.e. the probability of observing the computed chi2
        statistic (or an even higher value), given the null hypothesis
        that X _|_ Y | Zs.
    sufficient_data: bool
        A flag that indicates if the sample size is considered sufficient.
        As in [4], require at least 5 samples per parameter (on average).
        That is, the size of the data set must be greater than
        `5 * (c(X) - 1) * (c(Y) - 1) * prod([c(Z) for Z in Zs])`
        (c() denotes the variable cardinality).


    References
    ----------
    [1] Koller & Friedman, Probabilistic Graphical Models - Principles and Techniques, 2009
    Section 18.2.2.3 (page 789)
    [2] Neapolitan, Learning Bayesian Networks, Section 10.3 (page 600ff)
        http://www.cs.technion.ac.il/~dang/books/Learning%20Bayesian%20Networks(Neapolitan,%20Richard).pdf
    [3] Chi-square test https://en.wikipedia.org/wiki/Pearson%27s_chi-squared_test#Test_of_independence
    [4] Tsamardinos et al., The max-min hill-climbing BN structure learning algorithm, 2005, Section 4

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from pgmpy.estimators import ConstraintBasedEstimator
    >>> data = pd.DataFrame(np.random.randint(0, 2, size=(50000, 4)), columns=list('ABCD'))
    >>> data['E'] = data['A'] + data['B'] + data['C']
    >>> c = ConstraintBasedEstimator(data)
    >>> print(c.test_conditional_independence('A', 'C'))  # independent
    True
    >>> print(c.test_conditional_independence('A', 'B', 'D'))  # independent
    True
    >>> print(c.test_conditional_independence('A', 'B', ['D', 'E']))  # dependent
    False
    """

    if isinstance(Z, (frozenset, list, set, tuple)):
        Z = list(Z)
    else:
        Z = [Z]

    if "state_names" in kwargs.keys():
        state_names = kwargs["state_names"]
    else:
        state_names = {
            var_name: data.loc[:, var_name].unique() for var_name in data.columns
        }

    num_params = (
        (len(state_names[X]) - 1)
        * (len(state_names[Y]) - 1)
        * np.prod([len(state_names[z]) for z in Z])
    )
    sufficient_data = len(data) >= num_params * 5

    if not sufficient_data:
        warn(
            "Insufficient data for testing {0} _|_ {1} | {2}. ".format(X, Y, Z)
            + "At least {0} samples recommended, {1} present.".format(
                5 * num_params, len(data)
            )
        )

    # compute actual frequency/state_count table:
    # = P(X,Y,Zs)
    XYZ_state_counts = pd.crosstab(
        index=data[X], columns=[data[Y]] + [data[z] for z in Z]
    )
    # reindex to add missing rows & columns (if some values don't appear in data)
    row_index = state_names[X]
    column_index = pd.MultiIndex.from_product(
        [state_names[Y]] + [state_names[z] for z in Z], names=[Y] + Z
    )
    if not isinstance(XYZ_state_counts.columns, pd.MultiIndex):
        XYZ_state_counts.columns = pd.MultiIndex.from_arrays([XYZ_state_counts.columns])
    XYZ_state_counts = XYZ_state_counts.reindex(
        index=row_index, columns=column_index
    ).fillna(0)

    # compute the expected frequency/state_count table if X _|_ Y | Zs:
    # = P(X|Zs)*P(Y|Zs)*P(Zs) = P(X,Zs)*P(Y,Zs)/P(Zs)
    if Z:
        XZ_state_counts = XYZ_state_counts.sum(axis=1, level=Z)  # marginalize out Y
        YZ_state_counts = XYZ_state_counts.sum().unstack(Z)  # marginalize out X
    else:
        XZ_state_counts = XYZ_state_counts.sum(axis=1)
        YZ_state_counts = XYZ_state_counts.sum()
    Z_state_counts = YZ_state_counts.sum()  # marginalize out both

    XYZ_expected = pd.DataFrame(
        index=XYZ_state_counts.index, columns=XYZ_state_counts.columns
    )
    for X_val in XYZ_expected.index:
        if Z:
            for Y_val in XYZ_expected.columns.levels[0]:
                XYZ_expected.loc[X_val, Y_val] = (
                    XZ_state_counts.loc[X_val]
                    * YZ_state_counts.loc[Y_val]
                    / Z_state_counts
                ).values
        else:
            for Y_val in XYZ_expected.columns:
                XYZ_expected.loc[X_val, Y_val] = (
                    XZ_state_counts.loc[X_val]
                    * YZ_state_counts.loc[Y_val]
                    / float(Z_state_counts)
                )

    observed = XYZ_state_counts.values.flatten()
    expected = XYZ_expected.fillna(0).values.flatten()
    # remove elements where the expected value is 0;
    # this also corrects the degrees of freedom for chisquare
    observed, expected = zip(
        *((o, e) for o, e in zip(observed, expected) if not e == 0)
    )

    chi2, significance_level = stats.chisquare(observed, expected)

    return chi2, significance_level


def pearsonr(X, Y, Z, data):
    r"""
    Computes Pearson correlation coefficient and p-value for testing non-correlation. Should be used
    only on continuous data. In case when :math:`Z != \null` uses linear regression and computes pearson
    coefficient on residuals.

    Parameters
    ----------
    X: str
        The first variable for testing the independence condition X _|_ Y | Z

    Y: str
        The second variable for testing the independence condition X _|_ Y | Z

    Z: list/array-like
        A list of conditional variable for testing the condition X _|_ Y | Z

    data: pandas.DataFrame
        The dataset in which to test the indepenedence condition.

    Returns
    -------
    Pearson's correlation coefficient: float
    p-value: float

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
    [2] https://en.wikipedia.org/wiki/Partial_correlation#Using_linear_regression
    """
    # Step 1: Test if the inputs are correct
    if not hasattr(Z, "__iter__"):
        raise ValueError(
            "Variable Z. Expected type: iterable. Got type: {t}".format(t=type(Z))
        )
    else:
        Z = list(Z)

    if not isinstance(data, pd.DataFrame):
        raise ValueError(
            "Variable data. Expected type: pandas.DataFrame. Got type: {t}".format(
                t=type(data)
            )
        )

    # Step 2: If Z is empty compute a non-conditional test.
    if len(Z) == 0:
        return stats.pearsonr(data.loc[:, X], data.loc[:, Y])

    # Step 3: If Z is non-empty, use linear regression to compute residuals and test independence on it.
    else:
        X_coef = np.linalg.lstsq(data.loc[:, Z], data.loc[:, X], rcond=None)[0]
        Y_coef = np.linalg.lstsq(data.loc[:, Z], data.loc[:, Y], rcond=None)[0]

        residual_X = data.loc[:, X] - data.loc[:, Z].dot(X_coef)
        residual_Y = data.loc[:, Y] - data.loc[:, Z].dot(Y_coef)

        return stats.pearsonr(residual_X, residual_Y)



#!/usr/bin/env python

from itertools import chain

from io import BytesIO
import pyparsing as pp


# TODO input and output state


try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.ElementTree as etree
    except ImportError:
        # try:
        #    import xml.etree.cElementTree as etree
        #    commented out because xml.etree.cElementTree is giving errors with dictionary attributes
        print("Failed to import ElementTree from any known place")

import numpy as np

from pgmpy.models import BayesianModel
from pgmpy.factors.discrete import TabularCPD, State


class XMLBIFReader(object):
    """
    Base class for reading network file in XMLBIF format.
    """

    def __init__(self, path=None, string=None):
        """
        Initialisation of XMLBIFReader object.

        Parameters
        ----------
        path : file or str
            File of XMLBIF data
            File of XMLBIF data
        string : str
            String of XMLBIF data

        Examples
        --------
        # xmlbif_test.xml is the file present in
        # http://www.cs.cmu.edu/~fgcozman/Research/InterchangeFormat/
        >>> reader = XMLBIFReader("xmlbif_test.xml")
        """
        if path:
            self.network = etree.ElementTree(file=path).getroot().find("NETWORK")
        elif string:
            self.network = etree.fromstring(string.encode("utf-8")).find("NETWORK")
        else:
            raise ValueError("Must specify either path or string")
        self.network_name = self.network.find("NAME").text
        self.variables = self.get_variables()
        self.variable_parents = self.get_parents()
        self.edge_list = self.get_edges()
        self.variable_states = self.get_states()
        self.variable_CPD = self.get_values()
        self.variable_property = self.get_property()
        self.state_names = self.get_states()

    def get_variables(self):
        """
        Returns list of variables of the network

        Examples
        --------
        >>> reader = XMLBIF.XMLBIFReader("xmlbif_test.xml")
        >>> reader.get_variables()
        ['light-on', 'bowel-problem', 'dog-out', 'hear-bark', 'family-out']
        """
        variables = [
            variable.find("NAME").text for variable in self.network.findall("VARIABLE")
        ]
        return variables

    def get_edges(self):
        """
        Returns the edges of the network

        Examples
        --------
        >>> reader = XMLBIF.XMLBIFReader("xmlbif_test.xml")
        >>> reader.get_edges()
        [['family-out', 'light-on'],
         ['family-out', 'dog-out'],
         ['bowel-problem', 'dog-out'],
         ['dog-out', 'hear-bark']]
        """
        edge_list = [
            [value, key]
            for key in self.variable_parents
            for value in self.variable_parents[key]
        ]
        return edge_list

    def get_states(self):
        """
        Returns the states of variables present in the network

        Examples
        --------
        >>> reader = XMLBIF.XMLBIFReader("xmlbif_test.xml")
        >>> reader.get_states()
        {'bowel-problem': ['true', 'false'],
         'dog-out': ['true', 'false'],
         'family-out': ['true', 'false'],
         'hear-bark': ['true', 'false'],
         'light-on': ['true', 'false']}
        """
        variable_states = {
            variable.find("NAME").text: [
                outcome.text for outcome in variable.findall("OUTCOME")
            ]
            for variable in self.network.findall("VARIABLE")
        }
        return variable_states

    def get_parents(self):
        """
        Returns the parents of the variables present in the network

        Examples
        --------
        >>> reader = XMLBIF.XMLBIFReader("xmlbif_test.xml")
        >>> reader.get_parents()
        {'bowel-problem': [],
         'dog-out': ['family-out', 'bowel-problem'],
         'family-out': [],
         'hear-bark': ['dog-out'],
         'light-on': ['family-out']}
        """
        variable_parents = {
            definition.find("FOR").text: [
                edge.text for edge in definition.findall("GIVEN")
            ]
            for definition in self.network.findall("DEFINITION")
        }
        return variable_parents

    def get_values(self):
        """
        Returns the CPD of the variables present in the network

        Examples
        --------
        >>> reader = XMLBIF.XMLBIFReader("xmlbif_test.xml")
        >>> reader.get_values()
        {'bowel-problem': array([[ 0.01],
                                 [ 0.99]]),
         'dog-out': array([[ 0.99,  0.01,  0.97,  0.03],
                           [ 0.9 ,  0.1 ,  0.3 ,  0.7 ]]),
         'family-out': array([[ 0.15],
                              [ 0.85]]),
         'hear-bark': array([[ 0.7 ,  0.3 ],
                             [ 0.01,  0.99]]),
         'light-on': array([[ 0.6 ,  0.4 ],
                            [ 0.05,  0.95]])}
        """
        variable_CPD = {
            definition.find("FOR").text: list(map(float, table.text.split()))
            for definition in self.network.findall("DEFINITION")
            for table in definition.findall("TABLE")
        }
        for variable in variable_CPD:
            arr = np.array(variable_CPD[variable])
            arr = arr.reshape(
                (
                    len(self.variable_states[variable]),
                    arr.size // len(self.variable_states[variable]),
                ),
                order="F",
            )
            variable_CPD[variable] = arr
        return variable_CPD

    def get_property(self):
        """
        Returns the property of the variable

        Examples
        --------
        >>> reader = XMLBIF.XMLBIFReader("xmlbif_test.xml")
        >>> reader.get_property()
        {'bowel-problem': ['position = (190, 69)'],
         'dog-out': ['position = (155, 165)'],
         'family-out': ['position = (112, 69)'],
         'hear-bark': ['position = (154, 241)'],
         'light-on': ['position = (73, 165)']}
        """
        variable_property = {
            variable.find("NAME").text: [
                property.text for property in variable.findall("PROPERTY")
            ]
            for variable in self.network.findall("VARIABLE")
        }
        return variable_property

    def get_model(self):
        model = BayesianModel()
        model.add_nodes_from(self.variables)
        model.add_edges_from(self.edge_list)
        model.name = self.network_name

        tabular_cpds = []
        for var, values in self.variable_CPD.items():
            evidence_card = [
                len(self.variable_states[evidence_var])
                for evidence_var in self.variable_parents[var]
            ]
            cpd = TabularCPD(
                var,
                len(self.variable_states[var]),
                values,
                evidence=self.variable_parents[var],
                evidence_card=evidence_card,
                state_names={
                    var: self.state_names[var]
                    for var in chain([var], self.variable_parents[var])
                },
            )
            tabular_cpds.append(cpd)

        model.add_cpds(*tabular_cpds)

        for node, properties in self.variable_property.items():
            for prop in properties:
                if prop is not None:
                    prop_name, prop_value = map(lambda t: t.strip(), prop.split("="))
                    model.nodes[node][prop_name] = prop_value

        return model


class XMLBIFWriter(object):
    """
    Base class for writing XMLBIF network file format.
    """

    def __init__(self, model, encoding="utf-8", prettyprint=True):
        """
        Initialise a XMLBIFWriter object.

        Parameters
        ----------
        model: BayesianModel Instance
            Model to write
        encoding: str (optional)
            Encoding for text data
        prettyprint: Bool(optional)
            Indentation in output XML if true

        Examples
        --------
        >>> writer = XMLBIFWriter(model)
        """
        if not isinstance(model, BayesianModel):
            raise TypeError("model must an instance of BayesianModel")
        self.model = model

        self.encoding = encoding
        self.prettyprint = prettyprint

        self.xml = etree.Element("BIF", attrib={"VERSION": "0.3"})
        self.network = etree.SubElement(self.xml, "NETWORK")
        if self.model.name:
            etree.SubElement(self.network, "NAME").text = self.model.name
        else:
            etree.SubElement(self.network, "NAME").text = "UNTITLED"

        self.variables = self.get_variables()
        self.states = self.get_states()
        self.properties = self.get_properties()
        self.definition = self.get_definition()
        self.tables = self.get_values()

    def __str__(self):
        """
        Return the XML as string.
        """
        if self.prettyprint:
            self.indent(self.xml)
        f = BytesIO()
        et = etree.ElementTree(self.xml)
        et.write(f, encoding=self.encoding, xml_declaration=True)
        return f.getvalue().decode(self.encoding)

    def indent(self, elem, level=0):
        """
        Inplace prettyprint formatter.
        """
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self.indent(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    def get_variables(self):
        """
        Add variables to XMLBIF

        Return
        ------
        dict: dict of type {variable: variable tags}

        Examples
        --------
        >>> writer = XMLBIFWriter(model)
        >>> writer.get_variables()
        {'bowel-problem': <Element VARIABLE at 0x7fe28607dd88>,
         'family-out': <Element VARIABLE at 0x7fe28607de08>,
         'hear-bark': <Element VARIABLE at 0x7fe28607de48>,
         'dog-out': <Element VARIABLE at 0x7fe28607ddc8>,
         'light-on': <Element VARIABLE at 0x7fe28607de88>}
        """
        variables = self.model.nodes()
        variable_tag = {}
        for var in sorted(variables):
            variable_tag[var] = etree.SubElement(
                self.network, "VARIABLE", attrib={"TYPE": "nature"}
            )
            etree.SubElement(variable_tag[var], "NAME").text = var
        return variable_tag

    def get_states(self):
        """
        Add outcome to variables of XMLBIF

        Return
        ------
        dict: dict of type {variable: outcome tags}

        Examples
        --------
        >>> writer = XMLBIFWriter(model)
        >>> writer.get_states()
        {'dog-out': [<Element OUTCOME at 0x7ffbabfcdec8>, <Element OUTCOME at 0x7ffbabfcdf08>],
         'family-out': [<Element OUTCOME at 0x7ffbabfd4108>, <Element OUTCOME at 0x7ffbabfd4148>],
         'bowel-problem': [<Element OUTCOME at 0x7ffbabfd4088>, <Element OUTCOME at 0x7ffbabfd40c8>],
         'hear-bark': [<Element OUTCOME at 0x7ffbabfcdf48>, <Element OUTCOME at 0x7ffbabfcdf88>],
         'light-on': [<Element OUTCOME at 0x7ffbabfcdfc8>, <Element OUTCOME at 0x7ffbabfd4048>]}
        """
        outcome_tag = {}
        cpds = self.model.get_cpds()
        for cpd in cpds:
            var = cpd.variable
            outcome_tag[var] = []
            if cpd.state_names is None or cpd.state_names.get(var) is None:
                states = range(cpd.get_cardinality([var])[var])
            else:
                states = cpd.state_names[var]

            for state in states:
                state_tag = etree.SubElement(self.variables[var], "OUTCOME")
                state_tag.text = self._make_valid_state_name(state)
                outcome_tag[var].append(state_tag)
        return outcome_tag

    def _make_valid_state_name(self, state_name):
        """Transform the input state_name into a valid state in XMLBIF.
        XMLBIF states must start with a letter an only contain letters,
        numbers and underscores.
        """
        # TODO: Throw a warning that the state names are going to be modified instead of silently modifying it.
        s = str(state_name)
        s_fixed = (
            pp.CharsNotIn(pp.alphanums + "_")
            .setParseAction(pp.replaceWith("_"))
            .transformString(s)
        )
        if not s_fixed[0].isalpha():
            s_fixed = s_fixed
        return s_fixed

    def get_properties(self):
        """
        Add property to variables in XMLBIF

        Return
        ------
        dict: dict of type {variable: property tag}

        Examples
        --------
        >>> writer = XMLBIFWriter(model)
        >>> writer.get_property()
        {'light-on': <Element PROPERTY at 0x7f7a2ffac1c8>,
         'family-out': <Element PROPERTY at 0x7f7a2ffac148>,
         'hear-bark': <Element PROPERTY at 0x7f7a2ffac188>,
         'bowel-problem': <Element PROPERTY at 0x7f7a2ffac0c8>,
         'dog-out': <Element PROPERTY at 0x7f7a2ffac108>}
        """
        variables = self.model.nodes()
        property_tag = {}
        for var in sorted(variables):
            properties = self.model.nodes[var]
            property_tag[var] = etree.SubElement(self.variables[var], "PROPERTY")
            for prop, val in properties.items():
                property_tag[var].text = str(prop) + " = " + str(val)
        return property_tag

    def get_definition(self):
        """
        Add Definition to XMLBIF

        Return
        ------
        dict: dict of type {variable: definition tag}

        Examples
        --------
        >>> writer = XMLBIFWriter(model)
        >>> writer.get_definition()
        {'hear-bark': <Element DEFINITION at 0x7f1d48977408>,
         'family-out': <Element DEFINITION at 0x7f1d489773c8>,
         'dog-out': <Element DEFINITION at 0x7f1d48977388>,
         'bowel-problem': <Element DEFINITION at 0x7f1d48977348>,
         'light-on': <Element DEFINITION at 0x7f1d48977448>}
        """
        cpds = self.model.get_cpds()
        cpds.sort(key=lambda x: x.variable)
        definition_tag = {}
        for cpd in cpds:
            definition_tag[cpd.variable] = etree.SubElement(self.network, "DEFINITION")
            etree.SubElement(definition_tag[cpd.variable], "FOR").text = cpd.variable
            for parent in cpd.variables[1:]:
                etree.SubElement(definition_tag[cpd.variable], "GIVEN").text = parent

        return definition_tag

    def get_values(self):
        """
        Add Table to XMLBIF.

        Return
        ---------------
        dict: dict of type {variable: table tag}

        Examples
        -------
        >>> writer = XMLBIFWriter(model)
        >>> writer.get_values()
        {'dog-out': <Element TABLE at 0x7f240726f3c8>,
         'light-on': <Element TABLE at 0x7f240726f488>,
         'bowel-problem': <Element TABLE at 0x7f240726f388>,
         'family-out': <Element TABLE at 0x7f240726f408>,
         'hear-bark': <Element TABLE at 0x7f240726f448>}
        """
        cpds = self.model.get_cpds()
        definition_tag = self.definition
        table_tag = {}
        for cpd in cpds:
            table_tag[cpd.variable] = etree.SubElement(
                definition_tag[cpd.variable], "TABLE"
            )
            table_tag[cpd.variable].text = ""
            for val in cpd.get_values().ravel(order="F"):
                table_tag[cpd.variable].text += str(val) + " "

        return table_tag

    def write_xmlbif(self, filename):
        """
        Write the xml data into the file.

        Parameters
        ----------
        filename: Name of the file.

        Examples
        -------
        >>> writer = XMLBIFWriter(model)
        >>> writer.write_xmlbif(test_file)
        """
        with open(filename, "w") as fout:
            fout.write(self.__str__())



import unittest
import numpy.testing as np_test
import pandas as pd
import numpy as np

from pgmpy.factors.continuous import LinearGaussianCPD


class TestLGCPD(unittest.TestCase):
    # @unittest.skip("TODO")
    def test_class_init(self):
        mu = np.array([7, 13])
        sigma = np.array([[4, 3], [3, 6]])

        cpd1 = LinearGaussianCPD(
            "Y", evidence_mean=mu, evidence_variance=sigma, evidence=["X1", "X2"]
        )
        self.assertEqual(cpd1.variable, "Y")
        self.assertEqual(cpd1.evidence, ["X1", "X2"])

    def test_maximum_likelihood_estimator(self):
        # Obtain the X and Y which are jointly gaussian from the distribution
        # beta = [2, 0.7, 0.3]
        sigma_c = 4

        x_df = pd.read_csv("pgmpy/tests/test_factors/test_continuous/gbn_values_1.csv")

        mu = np.array([7, 13])
        sigma = np.array([[4, 3], [3, 6]])

        cpd1 = LinearGaussianCPD(
            "Y", evidence_mean=mu, evidence_variance=sigma, evidence=["X1", "X2"]
        )
        mean, variance = cpd1.fit(x_df, states=["(Y|X)", "X1", "X2"], estimator="MLE")
        np_test.assert_allclose(mean, [2.361152, 0.693147, 0.276383], rtol=1e-03)
        np_test.assert_allclose(variance, sigma_c, rtol=1e-1)

    @unittest.skip("TODO")
    def test_pdf(self):
        cpd1 = LinearGaussianCPD("x", [0.23], 0.56)
        cpd2 = LinearGaussianCPD("y", [0.67, 1, 4.56, 8], 2, ["x1", "x2", "x3"])
        np_test.assert_almost_equal(cpd1.assignment(1), 0.3139868)
        np_test.assert_almost_equal(cpd2.assignment(1, 1.2, 2.3, 3.4), 1.076e-162)

    @unittest.skip("TODO")
    def test_copy(self):
        cpd = LinearGaussianCPD("y", [0.67, 1, 4.56, 8], 2, ["x1", "x2", "x3"])
        copy = cpd.copy()

        self.assertEqual(cpd.variable, copy.variable)
        self.assertEqual(cpd.beta_0, copy.beta_0)
        self.assertEqual(cpd.variance, copy.variance)
        np_test.assert_array_equal(cpd.beta_vector, copy.beta_vector)
        self.assertEqual(cpd.evidence, copy.evidence)

        cpd.variable = "z"
        self.assertEqual(copy.variable, "y")
        cpd.variance = 0
        self.assertEqual(copy.variance, 2)
        cpd.beta_0 = 1
        self.assertEqual(copy.beta_0, 0.67)
        cpd.evidence = ["p", "q", "r"]
        self.assertEqual(copy.evidence, ["x1", "x2", "x3"])
        cpd.beta_vector = [2, 2, 2]
        np_test.assert_array_equal(copy.beta_vector, [1, 4.56, 8])

        copy = cpd.copy()

        copy.variable = "k"
        self.assertEqual(cpd.variable, "z")
        copy.variance = 0.3
        self.assertEqual(cpd.variance, 0)
        copy.beta_0 = 1.5
        self.assertEqual(cpd.beta_0, 1)
        copy.evidence = ["a", "b", "c"]
        self.assertEqual(cpd.evidence, ["p", "q", "r"])
        copy.beta_vector = [2.2, 2.2, 2.2]
        np_test.assert_array_equal(cpd.beta_vector, [2, 2, 2])

    @unittest.skip("TODO")
    def test_str(self):
        cpd1 = LinearGaussianCPD("x", [0.23], 0.56)
        cpd2 = LinearGaussianCPD("y", [0.67, 1, 4.56, 8], 2, ["x1", "x2", "x3"])
        self.assertEqual(cpd1.__str__(), "P(x) = N(0.23; 0.56)")
        self.assertEqual(
            cpd2.__str__(),
            "P(y | x1, x2, x3) = N(1.0*x1 + " "4.56*x2 + 8.0*x3 + 0.67; 2)",
        )


#    def test_mle_fit(self):
#        cpd = LinearGaussianCPD('Y',  [0.2, -2, 3, 7], 9.6, ['X1', 'X2', 'X3'])
#        gbn_values = pd.read_csv('gbn_values.csv')
#        cpd.fit(gbn_values)



# TODO: This variables being set in this file should move to setup.py


try:  # pragma: no cover
    import torch

    # Check if GPU is available
    if torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    dtype = torch.float
except ImportError:  # pragma: no cover
    torch = None
    device = None
    dtype = None


# This module initializes flags for optional dependencies
try:  # pragma: no cover
    import pandas

    HAS_PANDAS = True
except ImportError:  # pragma: no cover
    HAS_PANDAS = False
    pandas = None



#!/usr/bin/env python3

from collections import defaultdict
from itertools import chain

from pgmpy.models import BayesianModel
from pgmpy.models import MarkovModel
from pgmpy.models import FactorGraph
from pgmpy.models import JunctionTree
from pgmpy.models import DynamicBayesianNetwork
from pgmpy.factors.discrete import TabularCPD


class Inference(object):
    """
    Base class for all inference algorithms.

    Converts BayesianModel and MarkovModel to a uniform representation so that inference
    algorithms can be applied. Also it checks if all the associated CPDs / Factors are
    consistent with the model.

    Initialize inference for a model.

    Parameters
    ----------
    model: pgmpy.models.BayesianModel or pgmpy.models.MarkovModel or pgmpy.models.NoisyOrModel
        model for which to initialize the inference object.

    Examples
    --------
    >>> from pgmpy.inference import Inference
    >>> from pgmpy.models import BayesianModel
    >>> from pgmpy.factors.discrete import TabularCPD
    >>> student = BayesianModel([('diff', 'grade'), ('intel', 'grade')])
    >>> diff_cpd = TabularCPD('diff', 2, [[0.2, 0.8]])
    >>> intel_cpd = TabularCPD('intel', 2, [[0.3, 0.7]])
    >>> grade_cpd = TabularCPD('grade', 3, [[0.1, 0.1, 0.1, 0.1],
    ...                                     [0.1, 0.1, 0.1, 0.1],
    ...                                     [0.8, 0.8, 0.8, 0.8]],
    ...                        evidence=['diff', 'intel'], evidence_card=[2, 2])
    >>> student.add_cpds(diff_cpd, intel_cpd, grade_cpd)
    >>> model = Inference(student)

    >>> from pgmpy.models import MarkovModel
    >>> from pgmpy.factors import DiscreteFactor
    >>> import numpy as np
    >>> student = MarkovModel([('Alice', 'Bob'), ('Bob', 'Charles'),
    ...                        ('Charles', 'Debbie'), ('Debbie', 'Alice')])
    >>> factor_a_b = DiscreteFactor(['Alice', 'Bob'], cardinality=[2, 2], value=np.random.rand(4))
    >>> factor_b_c = DiscreteFactor(['Bob', 'Charles'], cardinality=[2, 2], value=np.random.rand(4))
    >>> factor_c_d = DiscreteFactor(['Charles', 'Debbie'], cardinality=[2, 2], value=np.random.rand(4))
    >>> factor_d_a = DiscreteFactor(['Debbie', 'Alice'], cardinality=[2, 2], value=np.random.rand(4))
    >>> student.add_factors(factor_a_b, factor_b_c, factor_c_d, factor_d_a)
    >>> model = Inference(student)
    """

    def __init__(self, model):
        self.model = model
        model.check_model()

        if isinstance(model, JunctionTree):
            self.variables = set(chain(*model.nodes()))
        else:
            self.variables = model.nodes()

        self.cardinality = {}
        self.factors = defaultdict(list)

        if isinstance(model, BayesianModel):
            for node in model.nodes():
                cpd = model.get_cpds(node)
                if isinstance(cpd, TabularCPD):
                    self.cardinality[node] = cpd.variable_card
                    cpd = cpd.to_factor()
                for var in cpd.scope():
                    self.factors[var].append(cpd)

        elif isinstance(model, (MarkovModel, FactorGraph, JunctionTree)):
            self.cardinality = model.get_cardinality()

            for factor in model.get_factors():
                for var in factor.variables:
                    self.factors[var].append(factor)

        elif isinstance(model, DynamicBayesianNetwork):
            self.start_bayesian_model = BayesianModel(model.get_intra_edges(0))
            self.start_bayesian_model.add_cpds(*model.get_cpds(time_slice=0))
            cpd_inter = [model.get_cpds(node) for node in model.get_interface_nodes(1)]
            self.interface_nodes = model.get_interface_nodes(0)
            self.one_and_half_model = BayesianModel(
                model.get_inter_edges() + model.get_intra_edges(1)
            )
            self.one_and_half_model.add_cpds(
                *(model.get_cpds(time_slice=1) + cpd_inter)
            )



import unittest
from mock import Mock, MagicMock, call
from pgmpy.estimators.ScoreCache import LRUCache, ScoreCache
from pgmpy.estimators import BicScore
import pandas as pd


class TestScoreCache(unittest.TestCase):
    def test_caching(self):
        function_mock = Mock(side_effect=[1, 2])
        cache = LRUCache(function_mock, max_size=2)

        self.assertEqual(cache("key1"), 1)
        self.assertEqual(cache("key2"), 2)

        # Test will fail if the cache calls the function mock again
        self.assertEqual(cache("key2"), 2)
        self.assertEqual(cache("key1"), 1)
        self.assertEqual(cache("key1"), 1)

    def test_small_cache(self):
        function_mock = Mock(side_effect=lambda key: {"key1": 1, "key2": 2}[key])
        cache = LRUCache(function_mock, max_size=1)

        cache("key1")
        cache("key1")  # cached
        cache("key1")  # cached
        cache("key2")
        cache("key2")  # cached
        cache("key1")

        expected_function_calls = [call("key1"), call("key2"), call("key1")]
        function_mock.assert_has_calls(expected_function_calls, any_order=False)

    def test_remove_least_recently_used(self):
        function_mock = Mock(
            side_effect=lambda key: {"key1": 1, "key2": 2, "key3": 3}[key]
        )
        cache = LRUCache(function_mock, max_size=2)

        cache("key1")
        cache("key2")
        cache("key3")  # kicks out 'key1'
        cache("key2")  # cached
        cache("key1")

        expected_function_calls = [
            call("key1"),
            call("key2"),
            call("key3"),
            call("key1"),
        ]
        function_mock.assert_has_calls(expected_function_calls, any_order=False)

    def test_score_cache_invalid_scorer(self):
        with self.assertRaises(AssertionError) as e:
            ScoreCache("invalid_scorer", None)

    def test_score_cache(self):
        def local_scores(*args):
            if args == ("key1", ["key2", "key3"]):
                return -1
            if args == ("key2", ["key3"]):
                return -2

            assert False, "Unhandled arguments"

        data = pd.DataFrame({"key1": [1, 2], "key2": [1, 2]})
        base_scorer = MagicMock(spec=BicScore)
        base_scorer.local_score = Mock(side_effect=local_scores)
        cache = ScoreCache(base_scorer, data)

        self.assertEqual(cache.local_score("key1", ["key2", "key3"]), -1)
        self.assertEqual(cache.local_score("key1", ["key2", "key3"]), -1)  # cached
        self.assertEqual(cache.local_score("key2", ["key3"]), -2)
        self.assertEqual(cache.local_score("key2", ["key3"]), -2)  # cached

        expected_function_calls = [
            call("key1", ["key2", "key3"]),
            call("key2", ["key3"]),
        ]
        base_scorer.local_score.assert_has_calls(
            expected_function_calls, any_order=False
        )



import unittest

import pandas as pd
import numpy as np
import networkx as nx

from pgmpy.estimators import ConstraintBasedEstimator
from pgmpy.independencies import Independencies
from pgmpy.models import BayesianModel
from pgmpy.base import DAG, UndirectedGraph


class TestConstraintBasedEstimator(unittest.TestCase):
    def test_build_skeleton(self):
        ind = Independencies(["B", "C"], ["A", ["B", "C"], "D"])
        ind = ind.closure()
        skel1, sep_sets1 = ConstraintBasedEstimator.build_skeleton("ABCD", ind)
        self.assertTrue(
            self._edge_list_equal(skel1.edges(), [("A", "D"), ("B", "D"), ("C", "D")])
        )

        sep_sets_ref1 = {
            frozenset({"A", "C"}): (),
            frozenset({"A", "B"}): (),
            frozenset({"C", "B"}): (),
        }
        self.assertEqual(sep_sets1, sep_sets_ref1)

        model = BayesianModel([("A", "C"), ("B", "C"), ("B", "D"), ("C", "E")])
        skel2, sep_sets2 = ConstraintBasedEstimator.build_skeleton(
            model.nodes(), model.get_independencies()
        )
        self.assertTrue(
            self._edge_list_equal(
                skel2, [("D", "B"), ("A", "C"), ("B", "C"), ("C", "E")]
            )
        )

        sep_sets_ref2 = {
            frozenset({"D", "C"}): ("B",),
            frozenset({"E", "B"}): ("C",),
            frozenset({"A", "D"}): (),
            frozenset({"E", "D"}): ("C",),
            frozenset({"E", "A"}): ("C",),
            frozenset({"A", "B"}): (),
        }
        # witnesses/seperators might change on each run, so we cannot compare directly
        self.assertEqual(sep_sets2.keys(), sep_sets_ref2.keys())
        self.assertEqual(
            [len(v) for v in sorted(sep_sets2.values())],
            [len(v) for v in sorted(sep_sets_ref2.values())],
        )

    def test_skeleton_to_pdag(self):
        data = pd.DataFrame(
            np.random.randint(0, 3, size=(1000, 3)), columns=list("ABD")
        )
        data["C"] = data["A"] - data["B"]
        data["D"] += data["A"]
        c = ConstraintBasedEstimator(data)
        pdag = c.skeleton_to_pdag(*c.estimate_skeleton())
        self.assertSetEqual(
            set(pdag.edges()), set([("B", "C"), ("A", "D"), ("A", "C"), ("D", "A")])
        )

        skel = UndirectedGraph([("A", "B"), ("A", "C")])
        sep_sets1 = {frozenset({"B", "C"}): ()}
        self.assertSetEqual(
            set(c.skeleton_to_pdag(skel, sep_sets1).edges()),
            set([("B", "A"), ("C", "A")]),
        )

        sep_sets2 = {frozenset({"B", "C"}): ("A",)}
        pdag2 = c.skeleton_to_pdag(skel, sep_sets2)
        self.assertSetEqual(
            set(c.skeleton_to_pdag(skel, sep_sets2).edges()),
            set([("A", "B"), ("B", "A"), ("A", "C"), ("C", "A")]),
        )

    def test_pdag_to_dag(self):
        pdag1 = nx.DiGraph(
            [("A", "B"), ("C", "B"), ("C", "D"), ("D", "C"), ("D", "A"), ("A", "D")]
        )
        dag1 = ConstraintBasedEstimator.pdag_to_dag(pdag1)
        self.assertTrue(
            ("A", "B") in dag1.edges()
            and ("C", "B") in dag1.edges()
            and len(dag1.edges()) == 4
        )

        pdag2 = nx.DiGraph([("B", "C"), ("D", "A"), ("A", "D"), ("A", "C")])
        dag2 = ConstraintBasedEstimator.pdag_to_dag(pdag2)
        self.assertTrue(
            set(dag2.edges()) == set([("B", "C"), ("A", "D"), ("A", "C")])
            or set(dag2.edges()) == set([("B", "C"), ("D", "A"), ("A", "C")])
        )

        pdag3 = nx.DiGraph([("B", "C"), ("D", "C"), ("C", "D"), ("A", "C")])
        dag3 = ConstraintBasedEstimator.pdag_to_dag(pdag3)
        self.assertSetEqual(
            set([("B", "C"), ("C", "D"), ("A", "C")]), set(dag3.edges())
        )

    def test_estimate_from_independencies(self):
        ind = Independencies(["B", "C"], ["A", ["B", "C"], "D"])
        ind = ind.closure()
        model = ConstraintBasedEstimator.estimate_from_independencies("ABCD", ind)

        self.assertSetEqual(
            set(model.edges()), set([("B", "D"), ("A", "D"), ("C", "D")])
        )

        model1 = BayesianModel([("A", "C"), ("B", "C"), ("B", "D"), ("C", "E")])
        model2 = ConstraintBasedEstimator.estimate_from_independencies(
            model1.nodes(), model1.get_independencies()
        )

        self.assertTrue(
            set(model2.edges()) == set(model1.edges())
            or set(model2.edges())
            == set([("B", "C"), ("A", "C"), ("C", "E"), ("D", "B")])
        )

    def test_estimate_skeleton(self):
        data = pd.DataFrame(
            np.random.randint(0, 2, size=(1000, 5)), columns=list("ABCDE")
        )
        data["F"] = data["A"] + data["B"] + data["C"]
        est = ConstraintBasedEstimator(data)
        skel, sep_sets = est.estimate_skeleton()
        self.assertTrue(
            self._edge_list_equal(skel.edges(), [("A", "F"), ("B", "F"), ("C", "F")])
        )

        sep_sets_ref = {
            frozenset({"D", "F"}): (),
            frozenset({"D", "B"}): (),
            frozenset({"A", "C"}): (),
            frozenset({"D", "E"}): (),
            frozenset({"E", "F"}): (),
            frozenset({"E", "C"}): (),
            frozenset({"E", "B"}): (),
            frozenset({"D", "C"}): (),
            frozenset({"A", "B"}): (),
            frozenset({"A", "E"}): (),
            frozenset({"B", "C"}): (),
            frozenset({"A", "D"}): (),
        }
        self.assertEqual(set(sep_sets.keys()), set(sep_sets_ref.keys()))

    def test_estimate_skeleton2(self):
        data = pd.DataFrame(
            np.random.randint(0, 2, size=(1000, 3)), columns=list("XYZ")
        )
        data["X"] += data["Z"]
        data["Y"] += data["Z"]
        est = ConstraintBasedEstimator(data)
        skel, sep_sets = est.estimate_skeleton()

        self.assertTrue(self._edge_list_equal(skel.edges(), [("X", "Z"), ("Y", "Z")]))
        self.assertEqual(sep_sets, {frozenset(("X", "Y")): ("Z",)})

    def test_estimate(self):
        data = pd.DataFrame(
            np.random.randint(0, 3, size=(1000, 3)), columns=list("XYZ")
        )
        data["sum"] = data.sum(axis=1)
        model = ConstraintBasedEstimator(data).estimate()
        self.assertSetEqual(
            set(model.edges()), set([("Z", "sum"), ("X", "sum"), ("Y", "sum")])
        )

    @staticmethod
    def _edge_list_equal(edges1, edges2):
        "Checks if two lists of undirected edges are equal."
        return (((X, Y) in edges2 or (Y, X) in edges2) for X, Y in edges1) and (
            ((X, Y) in edges1 or (Y, X) in edges1) for X, Y in edges2
        )
